# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)


# When using this file the name "Load" is what is imported
@declare_process_block_class("Dsi")
class dsiData(UnitModelBlockData):
    """
    Direct Steam Injection Unit Model

    This unit model is used to represent a direct steam injection
    process. There are no degrees of freedom, but the steam is mixed with the inlet fluid to heat it up.
    It is assumed that the pressure of the fluid doesn't change, i.e the steam loses its pressure.
    However, the enthalpy of the steam remains the same.
    This allows to use two different property packages for the steam and for the inlet fluid, however,
    it only works if the reference enthalpy of the steam and the inlet fluid are the same.

    It's basically a combination of a mixer and a translator.
    """

    # CONFIG are options for the unit model
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "steam_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
   **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "steam_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
        self.properties_milk_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )

        # We need to calculate the enthalpy of the composition, before adding additional enthalpy from the temperature difference.
        # so we'll add another state block to do that.
        tmp_dict["defined_state"] = False
        tmp_dict["has_phase_equilibrium"] = False
        self.properties_mixed_unheated = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of mixture, before accounting for temperature difference",
            **tmp_dict,
        )

        # Add outlet block
        tmp_dict["defined_state"] = False
        tmp_dict["has_phase_equilibrium"] = False
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict,
        )

        # Add steam inlet block
        steam_dict = dict(**self.config.steam_property_package_args)
        steam_dict["parameters"] = self.config.steam_property_package
        steam_dict["defined_state"] = True
        tmp_dict["has_phase_equilibrium"] = True

        self.properties_steam_in = self.config.steam_property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of steam inlet",
            **steam_dict,
        )

        # To calculate the amount of enthalpy to add to the inlet fluid, we need to know the difference in enthalpy between steam at that T and P
        # and steam at its inlet conditions. Note this is assuming that effects of composition (the steam will no longer be pure water) are negligible.
        # Note that this state block is just for calcuating, and not an actual inlet or outlet.

        steam_dict["defined_state"] = False  # This doesn't affect pure components.
        steam_dict["has_phase_equilibrium"] = True
        self.properties_steam_cooled = (
            self.config.steam_property_package.state_block_class(
                self.flowsheet().config.time,
                doc="Material properties of cooled steam",
                **steam_dict,
            )
        )

        # Add ports
        self.add_port(name="outlet", block=self.properties_out)
        self.add_port(name="inlet", block=self.properties_milk_in, doc="Inlet port")
        self.add_port(
            name="steam_inlet", block=self.properties_steam_in, doc="Steam inlet port"
        )

        # CONDITIONS

        # STEAM INTERMEDIATE BLOCK

        # Temperature (= other inlet temperature)
        @self.Constraint(
            self.flowsheet().time,
            doc="Set the temperature of the cooled steam to be the same as the inlet fluid",
        )
        def eq_steam_cooled_temperature(b, t):
            return (
                b.properties_steam_cooled[t].temperature
                == b.properties_milk_in[t].temperature
            )

        # Pressure (= other inlet pressure)
        @self.Constraint(
            self.flowsheet().time,
            doc="Set the pressure of the cooled steam to be the same as the inlet fluid",
        )
        def eq_steam_cooled_pressure(b, t):
            return (
                b.properties_steam_cooled[t].pressure
                == b.properties_milk_in[t].pressure
            )

        # Flow = steam_flow
        @self.Constraint(
            self.flowsheet().time,
            self.config.steam_property_package.component_list,
            doc="Set the composition of the cooled steam to be the same as the steam inlet",
        )
        def eq_steam_cooled_composition(b, t, c):
            return 0 == sum(
                b.properties_steam_cooled[t].get_material_flow_terms(p, c)
                - b.properties_steam_in[t].get_material_flow_terms(p, c)
                for p in b.properties_steam_in[t].phase_list
            )

        # CALCULATE ENTHALPY DIFFERENCE
        @self.Expression(
            self.flowsheet().time,
        )
        def steam_delta_h(b, t):
            """
            Calculate the difference in enthalpy between the steam inlet and the cooled steam.
            This is used to calculate the amount of enthalpy to add to the inlet fluid.
            """
            return (
                b.properties_steam_in[t].enth_mol
                - b.properties_steam_cooled[t].enth_mol
            ) * b.properties_steam_in[t].flow_mol

        # MIXING (without changing temperature)

        # Pressure (= inlet pressure)
        @self.Constraint(
            self.flowsheet().time,
            doc="Equivalent pressure balance",
        )
        def eq_mixed_pressure(b, t):
            return (
                b.properties_mixed_unheated[t].pressure
                == b.properties_milk_in[t].pressure
            )

        # Temperature (= inlet temperature)
        @self.Constraint(
            self.flowsheet().time,
            doc="Equivalent temperature balance",
        )
        def eq_mixed_temperature(b, t):
            return (
                b.properties_mixed_unheated[t].temperature
                == b.properties_milk_in[t].temperature
            )

        # Flow = inlet flow + steam flow
        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.component_list,
            doc="Mass balance",
        )
        def eq_mixed_composition(b, t, c):
            return 0 == sum(
                b.properties_milk_in[t].get_material_flow_terms(p, c)
                + (
                    b.properties_steam_in[t].get_material_flow_terms(p, c)
                    if c
                    in b.properties_steam_in[
                        t
                    ].component_list  # handle the case where a component isn't in the steam inlet (e.g no milk in helmholtz)
                    else 0
                )
                - b.properties_mixed_unheated[t].get_material_flow_terms(p, c)
                for p in b.properties_milk_in[t].phase_list
                if (p, c) in b.properties_milk_in[t].phase_component_set
            )  # handle the case where a component is not in that phase (e.g no milk vapor)

        # OUTLET BLOCK

        # Pressure (= inlet pressure)
        @self.Constraint(
            self.flowsheet().time,
            doc="Pressure balance",
        )
        def eq_outlet_pressure(b, t):
            return b.properties_out[t].pressure == b.properties_milk_in[t].pressure

        # Enthalpy (= mixed enthalpy + delta steam enthalpy)
        @self.Constraint(
            self.flowsheet().time,
            doc="Energy balance",
        )
        def eq_outlet_combined_enthalpy(b, t):
            return b.properties_out[t].enth_mol == b.properties_mixed_unheated[
                t
            ].enth_mol + (b.steam_delta_h[t] / b.properties_mixed_unheated[t].flow_mol)
        
        # Flow = mixed flow

        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.component_list,
            doc="Mass balance for the outlet",
        )
        def eq_outlet_composition(b, t, c):
            return 0 == sum(
                b.properties_out[t].get_material_flow_terms(p, c)
                - b.properties_mixed_unheated[t].get_material_flow_terms(p, c)
                for p in b.properties_out[t].phase_list
                if (p, c) in b.properties_out[t].phase_component_set
            )  # handle the case where a component is not in that phase (e.g no milk vapor)



    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

    def initialize(blk, *args, **kwargs):
        blk.properties_milk_in.initialize()
        blk.properties_steam_in.initialize()

        for t in blk.flowsheet().time:
            # copy temperature and pressure from properties_milk_in to properties_steam_cooled
            # blk.properties_steam_cooled[t].temperature.set_value(
            #     blk.properties_milk_in[t].temperature.value
            # )
            blk.properties_steam_cooled[t].pressure.set_value(
                blk.properties_milk_in[t].pressure.value
            )
            # Copy composition from properties_steam_in to properties_steam_cooled
            blk.properties_steam_cooled[t].flow_mol.set_value(
                blk.properties_steam_in[t].flow_mol.value
            )
            # If it's steam, there's only one component, so we prolly don't need to worry about composition.
            # But may want TODO this for other cases.

        blk.properties_steam_cooled.initialize()
        blk.properties_mixed_unheated.initialize()

        blk.properties_out.initialize()
        pass

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe(
                {
                    "outlet": self.outlet,
                    "inlet": self.inlet,
                    "steam_inlet": self.steam_inlet,
                },
                time_point=time_point,
            )
        except AttributeError:
            raise ConfigurationError(
                f"Unit model {self.name} does not have the standard Port "
                f"names (inlet and outlet). Please contact the unit model "
                f"developer to develop a unit specific stream table."
            )



from ahuora_builder.custom.thermal_utility_systems.steam_header import SteamHeader
from ahuora_builder.custom.custom_heater import DynamicHeater
import pytest
import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from property_packages.build_package import build_package
from pyomo.network import Arc, SequentialDecomposition
import idaes.logger as idaeslog



@pytest.fixture
def steam_header_base_case():
    def _make_case(num_inlets=2, num_outlets=3):
        # This defines the base case for all tests
        m = pyo.ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])
        # Create CustomHeader with 2 inlets and 3 outlets
        m.fs.header = SteamHeader(property_package=m.fs.water, num_inlets=num_inlets, num_outlets=num_outlets)

        # Set inlet conditions for inlet_1
        m.fs.header.mixer.inlet_1.flow_mol.fix(1000)  # mol/s
        m.fs.header.mixer.inlet_1.pressure.fix(201325)  # Pa
        m.fs.header.mixer.inlet_1.enth_mol.fix(m.fs.water.htpx((300 + 273.15) * pyo.units.K, m.fs.header.mixer.inlet_1.pressure[0]))  # J/mol

        # Initialise outlet flow for the let-down line (i.e., outlet_2 in this case)
        vent = m.fs.header.splitter.vent
        vent.flow_mol.fix(0)
        vent.flow_mol.unfix()
        return m
    return _make_case

def assert_header_solution(
        m,
        expected_total_flow,
        expected_vent_flow,
        expected_makeup_flow,
        expected_liquid_flow,
        has_heat_loss,
        has_pressure_loss,
    ):
    eps_rel = 1e-4 
    eps_abs = 1e-4
    # Check degrees of freedom
    assert degrees_of_freedom(m) == 0
    
    # Initialize the model
    m.fs.header.initialize()
    
    # Solve the model
    opt = pyo.SolverFactory("ipopt")
    results = opt.solve(m, tee=False)
    
    # Check that solve was successful
    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    
    # Verify mass balance: total inlet flow = total vapor outlet flow + liquid outlet flow
    total_inlet_flow = sum([pyo.value(inlet_sb[0].flow_mol) for inlet_sb in m.fs.header.inlet_blocks]) + pyo.value(m.fs.header.makeup_flow_mol[0])
    total_outlet_flow = sum([pyo.value(outlet_sb[0].flow_mol) for outlet_sb in m.fs.header.outlet_blocks])
    condensate_flow = pyo.value(m.fs.header.phase_separator.condensate_outlet.flow_mol[0])
    assert expected_total_flow == pytest.approx(total_outlet_flow + condensate_flow, rel=eps_rel, abs=eps_abs)
    assert expected_total_flow == pytest.approx(total_inlet_flow, rel=eps_rel, abs=eps_abs)
    
    assert expected_liquid_flow == pytest.approx(condensate_flow, rel=eps_rel, abs=eps_abs)
    assert expected_vent_flow == pytest.approx(pyo.value(m.fs.header.splitter.vent.flow_mol[0]), rel=eps_rel, abs=0.02)
    assert expected_makeup_flow == pytest.approx(pyo.value(m.fs.header.makeup_flow_mol[0]), rel=eps_rel, abs=0.02)

    # Verify pressure drop is applied correctly
    if pyo.value(m.fs.header.mixer.inlet_1.pressure[0]) > eps_rel:
        expected_min_inlet_pressure = pyo.value(m.fs.header.mixer.inlet_1.pressure[0])
    else:
        expected_min_inlet_pressure = pyo.value(m.fs.header.mixer.inlet_1.pressure[0])
    expected_outlet_pressure = expected_min_inlet_pressure + pyo.value(m.fs.header.deltaP[0])  # Accounts for pressure drop
    
    assert pyo.value(m.fs.header.splitter.outlet_1.pressure[0]) == pytest.approx(expected_outlet_pressure, rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.header.splitter.vent.pressure[0]) == pytest.approx(expected_outlet_pressure, rel=eps_rel, abs=eps_abs)

    assert pyo.value(m.fs.header.splitter.outlet_1.pressure[0]) == pytest.approx(pyo.value(m.fs.header.phase_separator.condensate_outlet.pressure[0]), rel=eps_rel, abs=eps_abs)
    assert pyo.value(m.fs.header.splitter.vent.pressure[0]) == pytest.approx(pyo.value(m.fs.header.phase_separator.condensate_outlet.pressure[0]), rel=eps_rel, abs=eps_abs)

    cooler_inlet_pressure = pyo.value(m.fs.header.mixer.mixed_state[0].pressure)
    cooler_outlet_pressure = pyo.value(m.fs.header.pressure[0])
    cooler_deltaP = -pyo.value(m.fs.header.deltaP[0])
    if total_outlet_flow > eps_abs:
        assert cooler_outlet_pressure == pytest.approx(cooler_inlet_pressure - cooler_deltaP, rel=eps_rel, abs=eps_abs)
    
    # Verify molar enthalpy relationships
    assert pyo.value(m.fs.header.splitter.outlet_1.enth_mol[0]) == pytest.approx(pyo.value(m.fs.header.splitter.vent.enth_mol[0]), rel=eps_rel, abs=eps_abs)

    hV = pyo.value(m.fs.header.phase_separator.vapor_outlet.enth_mol[0])  # J/mol
    hL = pyo.value(m.fs.header.phase_separator.condensate_outlet.enth_mol[0])  # J/mol
    if pyo.value(m.fs.header.phase_separator.condensate_outlet.flow_mol[0]) > eps_abs:
        assert hV >= hL

    # Verify cooling effect
    cooler_inlet_temperature = pyo.value(m.fs.header.mixer.mixed_state[0].temperature)
    cooler_outlet_temperature = pyo.value(m.fs.header.temperature[0])
    if total_outlet_flow > eps_abs:
        if has_heat_loss:
            assert cooler_outlet_temperature < cooler_inlet_temperature
        else:
            assert cooler_outlet_temperature == pytest.approx(cooler_inlet_temperature, rel=eps_rel, abs=eps_abs)

    cooler_inlet_total_energy = pyo.value(m.fs.header.mixer.mixed_state[0].enth_mol * m.fs.header.mixer.mixed_state[0].flow_mol)
    cooler_outlet_total_energy = pyo.value(m.fs.header.enth_mol[0] * m.fs.header.total_flow_mol[0])
    cooler_duty = -pyo.value(m.fs.header.heat_duty[0])
    assert cooler_outlet_total_energy == pytest.approx(cooler_inlet_total_energy - cooler_duty, rel=eps_rel, abs=eps_abs)

    # Test that the unit has the expected structure
    assert hasattr(m.fs.header, 'mixer')
    assert hasattr(m.fs.header, 'cooler')
    assert hasattr(m.fs.header, 'phase_separator')
    assert hasattr(m.fs.header, 'splitter')
    assert hasattr(m.fs.header.phase_separator, 'condensate_outlet')
    
    # Test that ports exist
    assert hasattr(m.fs.header.mixer, 'inlet_1')
    assert hasattr(m.fs.header.splitter, 'outlet_1')
    assert hasattr(m.fs.header.splitter, 'vent')

    inlet_mass_list = [pyo.value(i[0].flow_mass) for i in m.fs.header.inlet_blocks] + [ pyo.value(m.fs.header.makeup_flow_mol[0] * 0.01801528)]  # convert mol/s to kg/s
    outlet_mass_list = [pyo.value(i[0].flow_mass) for i in m.fs.header.outlet_blocks]
    condensate_mass = pyo.value(m.fs.header.phase_separator.condensate_outlet_state[0].flow_mass)
    # Test all mass flows are non-zero and sum of inlet and outlet masses is approx same
    assert sum(inlet_mass_list) == pytest.approx(sum(outlet_mass_list) + condensate_mass, rel=eps_rel, abs=eps_abs)
    
    min_inlet_pressure = min([pyo.value(i[0].pressure) for i in m.fs.header.inlet_blocks])
    min_outlet_pressure = min([pyo.value(i[0].pressure) for i in m.fs.header.outlet_blocks])
    min_outlet_pressure = min(pyo.value(m.fs.header.phase_separator.condensate_outlet.pressure[0]),  min_outlet_pressure)
    
    # Test minimum input pressure is >= outlet pressure and all pressure are non-zero
    assert min_inlet_pressure >= min_outlet_pressure - eps_abs
    assert min_inlet_pressure > eps_abs
    assert min_outlet_pressure > eps_abs

def test_header_with_inlet_greater_than_outlet_flow(steam_header_base_case):
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    m = steam_header_base_case(num_inlets=1, num_outlets=1)
    
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(800)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(-1000)  # W -> heat loss
    m.fs.header.deltaP[0].fix(-1000)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1000,
        expected_vent_flow=200,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_condensate_in_inlet_flow(steam_header_base_case):
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    m = steam_header_base_case(num_inlets=1, num_outlets=1)

    h_val = pyo.value(m.fs.water.htpx(p=201325 * pyo.units.Pa, x=0.9))
    m.fs.header.mixer.inlet_1.enth_mol.fix(h_val)  # J/mol

    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(800)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1000,
        expected_vent_flow=100,
        expected_makeup_flow=0,
        expected_liquid_flow=100,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_inlet_equal_to_outlet_flow(steam_header_base_case):
    # Case where there is zero liquid flow and the header is balanced
    m = steam_header_base_case(num_inlets=1, num_outlets=1)
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(1000)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1000,
        expected_vent_flow=0,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_inlet_less_than_outlet_flow(steam_header_base_case):
    # Case where there is zero liquid flow and the header is balanced
    m = steam_header_base_case(num_inlets=1, num_outlets=1)
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(1200)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1200,
        expected_vent_flow=0,
        expected_makeup_flow=200,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_zero_flow(steam_header_base_case):
    # Case where there is zero liquid flow and the header is balanced
    m = steam_header_base_case(num_inlets=1, num_outlets=1)
    m.fs.header.mixer.inlet_1.flow_mol.fix(0)  # mol/s
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(0)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=0,
        expected_vent_flow=0,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_inlet_greater_than_three_defined_outlet_flows(steam_header_base_case):
    # Case where there is zero liquid flow and there is more known inlet than outlet flows
    m = steam_header_base_case(num_inlets=1, num_outlets=3)
    
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(500)  # mol/s
    m.fs.header.splitter.outlet_2.flow_mol.fix(100)  # mol/s
    m.fs.header.splitter.outlet_3.flow_mol.fix(200)  # mol/s

    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(-1000)  # W -> heat loss
    m.fs.header.deltaP[0].fix(-1000)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1000,
        expected_vent_flow=200,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_inlet_less_than_two_defined_outlet_flows(steam_header_base_case):
    # Case where there is zero liquid flow and the header is balanced
    m = steam_header_base_case(num_inlets=1, num_outlets=2)
    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(600)  # mol/s
    m.fs.header.splitter.outlet_2.flow_mol.fix(600)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1200,
        expected_vent_flow=0,
        expected_makeup_flow=200,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_header_with_two_defined_inlet_less_than_two_defined_outlet_flows(steam_header_base_case):
    # Case where there is zero liquid flow and the header is balanced
    m = steam_header_base_case(num_inlets=2, num_outlets=2)
    m.fs.header.mixer.inlet_2.flow_mol.fix(100)  # mol/s
    m.fs.header.mixer.inlet_2.pressure.fix(221325)  # Pa
    m.fs.header.mixer.inlet_2.enth_mol.fix(m.fs.water.htpx((310 + 273.15) * pyo.units.K, m.fs.header.mixer.inlet_1.pressure[0]))  # J/mol

    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(600)  # mol/s
    m.fs.header.splitter.outlet_2.flow_mol.fix(600)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    assert_header_solution(
        m,
        expected_total_flow=1200,
        expected_vent_flow=0,
        expected_makeup_flow=100,
        expected_liquid_flow=0,
        has_heat_loss=True if pyo.value(m.fs.header.heat_duty[0]) < 0 else False,
        has_pressure_loss=True if pyo.value(m.fs.header.deltaP[0]) < 0 else False,
    )

def test_sequential_decomposition(steam_header_base_case):
    m = steam_header_base_case(num_inlets=1, num_outlets=1)

    m.fs.heater = DynamicHeater(property_package=m.fs.water)

    # Set user steam demand
    m.fs.header.splitter.outlet_1.flow_mol.fix(600)  # mol/s
    # Set cooler specifications
    m.fs.header.heat_duty[0].fix(0)  # W -> heat loss
    m.fs.header.deltaP[0].fix(0)  # Pa -> pressure loss

    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.inlet.enth_mol.fix(m.fs.water.htpx((300 + 273.15) * pyo.units.K, m.fs.header.mixer.inlet_1.pressure[0]))  # J/mol
    m.fs.heater.inlet.pressure.fix(101325)

    m.fs.heater.heat_duty.fix(0)

    m.fs.header.mixer.inlet_1.flow_mol.unfix()
    m.fs.header.mixer.inlet_1.enth_mol.unfix()
    m.fs.header.mixer.inlet_1.pressure.unfix()

    m.fs.heater_to_header = Arc(
        source=m.fs.heater.outlet, destination=m.fs.header.mixer.inlet_1
    )

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    assert degrees_of_freedom(m) == 0

    # create Sequential Decomposition object
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    # create computation graph
    G = seq.create_graph(m)
    heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # get calculation order
    order = seq.calculation_order(G)

    for o in heuristic_tear_set:
        print(o.name)

    print("Initialization order:")
    for o in order:
        print(o[0].name)

    # define unit initialisation function
    def init_unit(unit):
        unit.initialize()

    # run sequential decomposition
    seq.run(m, init_unit)

def test_header_with_inlet_steam_generator():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])

    # Add a heater before the header
    m.fs.boiler = DynamicHeater(
        property_package=m.fs.water,
        has_pressure_change=True,
    )
    # Create CustomHeader with 2 inlets and 3 outlets
    m.fs.header = SteamHeader(
        property_package=m.fs.water, 
        num_inlets=1, 
        num_outlets=1
    )
    # Connect heater outlet to header inlet_1
    m.boiler_to_header_arc = Arc(
        source=m.fs.boiler.outlet, destination=m.fs.header.mixer.inlet_1
    )
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet conditions for inlet_1
    m.fs.boiler.inlet.flow_mol.fix(1000)  # mol/s
    m.fs.boiler.inlet.pressure.fix(200 * 1000)  # Pa
    m.fs.boiler.inlet.enth_mol.fix(
        m.fs.water.htpx(
            (15 + 273.15) * pyo.units.K, 
            m.fs.boiler.inlet.pressure[0]
        )
    )  # J/mol
    m.fs.boiler.heat_duty.fix(
        50.577 * 1000 * 1000
    )
    m.fs.boiler.deltaP.fix(
        0.0
    )
    m.fs.header.splitter.outlet_1.flow_mol.fix(500)
    m.fs.header.heat_duty.fix(
        0.0
    )
    m.fs.header.deltaP.fix(
        0.0
    )
    
    # Initialize the heater first
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    # create computation graph
    G = seq.create_graph(m)
    heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # get calculation order
    order = seq.calculation_order(G)

    for o in heuristic_tear_set:
        print(o.name)

    for o in order:
        print(o[0].name)

    # define unit initialisation function
    def init_unit(unit):
        unit.initialize()

    # run sequential decomposition
    seq.run(m, init_unit) 

    dof = degrees_of_freedom(m)

    opt = pyo.SolverFactory("ipopt_v2")
    results = opt.solve(m, tee=False)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal       

    assert_header_solution(
        m=m, 
        expected_total_flow=1000,
        expected_vent_flow=500,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=0,
        has_pressure_loss=0,
    )

def test_header_with_inlet_steam_generator_with_insufficent_flow():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])

    # Add a heater before the header
    m.fs.boiler = DynamicHeater(
        property_package=m.fs.water,
        has_pressure_change=True,
    )
    # Create CustomHeader with 2 inlets and 3 outlets
    m.fs.header = SteamHeader(
        property_package=m.fs.water, 
        num_inlets=1, 
        num_outlets=1
    )
    # Connect heater outlet to header inlet_1
    m.boiler_to_header_arc = Arc(
        source=m.fs.boiler.outlet, destination=m.fs.header.mixer.inlet_1
    )
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet conditions for inlet_1
    m.fs.boiler.inlet.flow_mol.fix(1000)  # mol/s
    m.fs.boiler.inlet.pressure.fix(200 * 1000)  # Pa
    m.fs.boiler.inlet.enth_mol.fix(
        m.fs.water.htpx(
            (15 + 273.15) * pyo.units.K, 
            m.fs.boiler.inlet.pressure[0]
        )
    )  # J/mol
    m.fs.boiler.heat_duty.fix(
        50.577 * 1000 * 1000
    )
    m.fs.boiler.deltaP.fix(
        0.0
    )
    m.fs.header.splitter.outlet_1.flow_mol.fix(1500)
    m.fs.header.heat_duty.fix(
        0.0
    )
    m.fs.header.deltaP.fix(
        0.0
    )
    
    # Initialize the heater first
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    # create computation graph
    G = seq.create_graph(m)
    heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # get calculation order
    order = seq.calculation_order(G)

    for o in heuristic_tear_set:
        print(o.name)

    for o in order:
        print(o[0].name)

    # define unit initialisation function
    def init_unit(unit):
        unit.initialize()

    # run sequential decomposition
    seq.run(m, init_unit) 

    opt = pyo.SolverFactory("ipopt_v2")
    results = opt.solve(m, tee=False)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal       
    
    assert_header_solution(
        m=m, 
        expected_total_flow=1500,
        expected_vent_flow=0,
        expected_makeup_flow=500,
        expected_liquid_flow=0,
        has_heat_loss=0,
        has_pressure_loss=0,
    )
 
def test_header_with_inlet_steam_generator_with_losses():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])

    # Add a heater before the header
    m.fs.boiler = DynamicHeater(
        property_package=m.fs.water,
        has_pressure_change=True,
    )
    # Create CustomHeader with 2 inlets and 3 outlets
    m.fs.header = SteamHeader(
        property_package=m.fs.water, 
        num_inlets=1, 
        num_outlets=1
    )
    # Connect heater outlet to header inlet_1
    m.boiler_to_header_arc = Arc(
        source=m.fs.boiler.outlet, destination=m.fs.header.mixer.inlet_1
    )
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet conditions for inlet_1
    m.fs.boiler.inlet.flow_mol.fix(1000)  # mol/s
    m.fs.boiler.inlet.pressure.fix(200 * 1000)  # Pa
    m.fs.boiler.inlet.enth_mol.fix(
        m.fs.water.htpx(
            (15 + 273.15) * pyo.units.K, 
            m.fs.boiler.inlet.pressure[0]
        )
    )  # J/mol
    m.fs.boiler.heat_duty.fix(
        50.577 * 1000 * 1000
    )
    m.fs.boiler.deltaP.fix(
        0.0
    )
    m.fs.header.splitter.outlet_1.flow_mol.fix(500)
    m.fs.header.heat_duty.fix(
        -10.0
    )
    m.fs.header.deltaP.fix(
        -5000.0
    )
    
    # Initialize the heater first
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    # create computation graph
    G = seq.create_graph(m)
    heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # get calculation order
    order = seq.calculation_order(G)

    for o in heuristic_tear_set:
        print(o.name)

    for o in order:
        print(o[0].name)

    # define unit initialisation function
    def init_unit(unit):
        unit.initialize()

    # run sequential decomposition
    seq.run(m, init_unit) 

    opt = pyo.SolverFactory("ipopt_v2")
    results = opt.solve(m, tee=False)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal       

    assert_header_solution(
        m=m, 
        expected_total_flow=1000,
        expected_vent_flow=500,
        expected_makeup_flow=0,
        expected_liquid_flow=0,
        has_heat_loss=True,
        has_pressure_loss=True,
    )
 
def test_header_with_inlet_steam_generator_with_wet_steam():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])

    # Add a heater before the header
    m.fs.boiler = DynamicHeater(
        property_package=m.fs.water,
        has_pressure_change=True,
    )
    # Create CustomHeader with 2 inlets and 3 outlets
    m.fs.header = SteamHeader(
        property_package=m.fs.water, 
        num_inlets=1, 
        num_outlets=1
    )
    # Connect heater outlet to header inlet_1
    m.boiler_to_header_arc = Arc(
        source=m.fs.boiler.outlet, destination=m.fs.header.mixer.inlet_1
    )
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet conditions for inlet_1
    m.fs.boiler.inlet.flow_mol.fix(1000)  # mol/s
    m.fs.boiler.inlet.pressure.fix(200 * 1000)  # Pa
    m.fs.boiler.inlet.enth_mol.fix(
        m.fs.water.htpx(
            (15 + 273.15) * pyo.units.K, 
            m.fs.boiler.inlet.pressure[0]
        )
    )  # J/mol
    m.fs.boiler.heat_duty.fix(
        27784964.63
    )
    m.fs.boiler.deltaP.fix(
        0.0
    )
    m.fs.header.splitter.outlet_1.flow_mol.fix(500)
    m.fs.header.heat_duty.fix(
        0.0
    )
    m.fs.header.deltaP.fix(
        0.0
    )
    
    # Initialize the heater first
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    # create computation graph
    G = seq.create_graph(m)
    heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # get calculation order
    order = seq.calculation_order(G)

    for o in heuristic_tear_set:
        print(o.name)

    for o in order:
        print(o[0].name)

    # define unit initialisation function
    def init_unit(unit):
        unit.initialize()

    # run sequential decomposition
    seq.run(m, init_unit) 

    opt = pyo.SolverFactory("ipopt_v2")
    results = opt.solve(m, tee=False)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal       

    assert_header_solution(
        m=m, 
        expected_total_flow=1000,
        expected_vent_flow=0,
        expected_makeup_flow=0.0,
        expected_liquid_flow=500,
        has_heat_loss=0,
        has_pressure_loss=0,
    )





# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError
from idaes.models.unit_models.translator import TranslatorData
# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from enum import Enum

# Set up logger
_log = idaeslog.getLogger(__name__)

class TranslatorType(Enum):
    """
    Enum of supported translator types. This allows you to change what variables are translated across.
    """
    pressure_enthalpy = "pressure_enthalpy"
    pressure_vapor_fraction = "pressure_vapor_fraction"
    pressure_temperature = "pressure_temperature"


# When using this file the name "GenericTranslator" is what is imported
@declare_process_block_class("GenericTranslator")
class GenericTranslatorData(TranslatorData):
    """
    GenericTranslator.

    This is used to translate between two different property packages, and supports dropping compounds that are not present.

    For example, if you have a stream of water/milk, and it's almost all water, this allows you to translate the stream to a water-only stream.

    It works by fixing the temperature and pressure, and flow of each component in the outlet stream to be the same as the inlet stream.
     
    """
    CONFIG = TranslatorData.CONFIG()

    CONFIG.declare(
        "translator_type",
        ConfigValue(
            default=TranslatorType.pressure_enthalpy.value,
            description="Translator type to use for translating properties",
            doc="""
            Depending on the property packages you are using and the phases that are present, it may make sense to use different translators.
            This allows you to select what variables are translated across.
            """,
        ),
    )


    def build(self):
        self.CONFIG.outlet_state_defined = False # See constraint for flow
        #self.CONFIG.has_phase_equilibrium = True # I don't think it matters if this is set, becuase in theory the phase equilibrium should
        # already have been calculated in the inlet stream.
        super().build()

        # Pressure (= inlet pressure)
        @self.Constraint(
            self.flowsheet().time,
            doc="Pressure balance",
        )
        def eq_outlet_pressure(b, t):
            return b.properties_in[t].pressure == b.properties_out[t].pressure

        # Flow
        @self.Constraint(
            self.flowsheet().time,
            self.config.outlet_property_package.component_list,
            doc="Mass balance for the outlet",
        )
        def eq_outlet_composition(b, t, c):
            return 0 == sum(
                b.properties_out[t].get_material_flow_terms(p, c)
                - b.properties_in[t].get_material_flow_terms(p, c)
                for p in b.properties_out[t].phase_list
                if (p, c) in b.properties_out[t].phase_component_set
            ) 

        if self.config.translator_type == TranslatorType.pressure_enthalpy.value:
            # Enthalpy (= inlet enthalpy)
            @self.Constraint(
                self.flowsheet().time,
                doc="Enthalpy balance",
            )
            def eq_outlet_enth_mol(b, t):
                return (
                    b.properties_in[t].enth_mol == b.properties_out[t].enth_mol
                )
        elif self.config.translator_type == TranslatorType.pressure_vapor_fraction.value:
            # Vapor fraction (= inlet vapor fraction)
            # TODO: We might be able to make this smoother, by extending the vapor fraction
            # below 0 and above 1 to make solving more reliable. See
            # https://github.com/waikato-ahuora-smart-energy-systems/PropertyPackages/blob/8c6ee67b9d028ba0fdd1c937d9dcda821595b7d1/property_packages/helmholtz/helmholtz_extended.py#L104
            @self.Constraint(
                self.flowsheet().time,
                doc="Vapor fraction balance",
            )
            def eq_outlet_vapor_frac(b, t):
                return (
                    b.properties_in[t].vapor_frac == b.properties_out[t].vapor_frac
                )
        elif self.config.translator_type == TranslatorType.pressure_temperature.value:
            # Temperature (= inlet temperature)
            @self.Constraint(
                self.flowsheet().time,
                doc="Temperature balance",
            )
            def eq_outlet_temperature(b, t):
                return (
                    b.properties_in[t].temperature == b.properties_out[t].temperature
                )
        else:
            raise ConfigurationError(
                f"Translator type {self.CONFIG.translator_type} is not supported."
            )
        
        




import pytest

from ahuora_builder.custom.thermal_utility_systems.simple_heat_pump import (
    SimpleHeatPump,
)

import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
#from property_packages.build_package import build_package # build package still broken for windows users
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    PhaseType,
    StateVars,
    AmountBasis,
    )

from pyomo.network import Arc, SequentialDecomposition
import idaes.logger as idaeslog
import math

def _make_case():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic = False)
    m.fs.water = HelmholtzParameterBlock(
                    pure_component="h2o",
                    phase_presentation=PhaseType.LG,
                    state_vars=StateVars.PH,
                    amount_basis=AmountBasis.MOLE,
                    )
    #build_package("helmholtz", ["water"], ["Liq", "Vap"])

    m.fs.hp = SimpleHeatPump(source={
        "property_package": m.fs.water,
        "has_pressure_change": False,
    }, sink={
        "property_package": m.fs.water,
        "has_pressure_change": False,
    })
    
    # Set source (hot side) conditions
    T_source_in = (80 + 273.15) * pyo.units.K  # K
    P_source_in = 4e5 * pyo.units.Pa  # Pa
    F_source_in = 29 * 15.4 * pyo.units.mol / pyo.units.s  # mol/s

    m.fs.hp.source_inlet.flow_mol.fix(F_source_in)  # mol/s
    m.fs.hp.source_inlet.pressure.fix(P_source_in)  # 4 Bar
    m.fs.hp.source_inlet.enth_mol.fix(
        m.fs.water.htpx(T_source_in , m.fs.hp.source_inlet.pressure[0])
    )  # J/mol
    m.fs.hp.approach_temperature.fix(10)  # K

    # Set sink (cold side) conditions
    T_sink_in = (90 + 273.15) * pyo.units.K  # K
    P_sink_in = 8e5 * pyo.units.Pa  # Pa
    F_sink_in = 1.1 * 15.4 * pyo.units.mol / pyo.units.s  # mol/s

    m.fs.hp.sink_inlet.flow_mol.fix(F_sink_in)  # mol/s
    m.fs.hp.sink_inlet.pressure.fix(P_sink_in)  # 10 Bar
    m.fs.hp.sink_inlet.enth_mol.fix(
        m.fs.water.htpx(T_sink_in, m.fs.hp.sink_inlet.pressure[0])
    )  # J/mol
    
    return m
 
def _assert_solution(m):
    eps_rel = 1e-4 
    eps_abs = 0.1
    m.fs.hp.initialize()

    assert degrees_of_freedom(m) == 0

    solver = pyo.SolverFactory("ipopt")
    solver.solve(m, tee=True)
    
    # Check heat balances Qsink = Qsource + W
    Q_sink = pyo.value(m.fs.hp.sink_inlet.flow_mol[0] * (m.fs.hp.sink_outlet.enth_mol[0] - m.fs.hp.sink_inlet.enth_mol[0] ))
    Q_source = pyo.value(m.fs.hp.source_inlet.flow_mol[0] * (m.fs.hp.source_inlet.enth_mol[0] - m.fs.hp.source_outlet.enth_mol[0] ))
    
    assert Q_sink - pyo.value(m.fs.hp.work_mechanical[0]) - Q_source == pytest.approx(0, rel=eps_rel, abs=eps_abs)     
    assert abs(pyo.value(m.fs.hp.source.heat[0])) == pytest.approx(Q_source, rel=eps_rel, abs=eps_abs)
    assert abs(pyo.value(m.fs.hp.sink.heat[0])) == pytest.approx(Q_sink, rel=eps_rel, abs=eps_abs)
    
    # Check that CoP is calculated correctly
    assert pyo.value(m.fs.hp.coefficient_of_performance) == pytest.approx(Q_sink / pyo.value(m.fs.hp.work_mechanical[0]), rel=eps_rel, abs=eps_abs)
    
    # Check that duties and CoP are as expected from datasheet
    assert pyo.value(m.fs.hp.coefficient_of_performance) == pytest.approx(1.9, rel=0.05, abs=0.2)
    assert pyo.value(m.fs.hp.sink.heat[0]) == pytest.approx(720e3, rel=0.05, abs=50e3)
    
    m.fs.hp.report()
    

def test_heat_pump_with_work_and_efficiency():
    m = _make_case()
    # Fix efficiency to the Ovolondo heat pump
    m.fs._ = pyo.Constraint(expr=m.fs.hp.efficiency == 0.3305 + 0.2723 * pyo.tanh((m.fs.hp.delta_temperature_lift[0] - 57.64) / 74.19))
    m.fs.hp.approach_temperature.fix(10)  # K
    # Fix mechanical work to the example
    m.fs.hp.work_mechanical.fix(380e3)  # W
    
    assert degrees_of_freedom(m) == 0
    
    _assert_solution(m)
    
def test_heat_pump_with_sink_outlet_and_efficiency():
    m = _make_case()
    
    # Fix efficiency to the Ovolondo heat pump
    m.fs._ = pyo.Constraint(expr=m.fs.hp.efficiency == 0.3305 + 0.2723 * pyo.tanh((m.fs.hp.delta_temperature_lift[0] - 57.64) / 74.19))
    m.fs.hp.approach_temperature.fix(10)  # K
    # Fix sink outlet temperature to the Ovolondo heat pump
    T_sink_out = (171 + 273.15) * pyo.units.K  # K
    m.fs.hp.sink_outlet.enth_mol.fix(
        m.fs.water.htpx(T_sink_out , m.fs.hp.sink_inlet.pressure[0])
    )  # J/mol
    
    assert degrees_of_freedom(m) == 0
    
    _assert_solution(m)





# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    Param,
    log,
    Reference,
    PositiveReals,
    ExternalFunction,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
)

import idaes.logger as idaeslog
from idaes.core.util.functions import functions_lib
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.heater import (
    _make_heater_config_block,
    _make_heater_control_volume,
)

from idaes.core.util.misc import add_object_reference
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.initialization import SingleControlVolumeUnitInitializer
from idaes.models.unit_models.heat_exchanger import HX0DInitializer, _make_heat_exchanger_config, HeatExchangerData
from .inverted import add_inverted, initialise_inverted
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("CustomHeatExchanger", doc="Simple 0D heat exchanger model.")
class CustomHeatExchangerData(HeatExchangerData):

    def build(self,*args,**kwargs) -> None:
        """
        Begin building model.
        """
        super().build(*args,**kwargs)
        # Add an inverted DeltaP
        add_inverted(self.hot_side, "deltaP")
        add_inverted(self.cold_side, "deltaP")

    def initialize_build(
        self,
        state_args_1=None,
        state_args_2=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Heat exchanger initialization method.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for the hot side (see documentation of the specific
                property package) (default = {}).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for the cold side (see documentation of the specific
                property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units), (default
                = (1000 J/s))

        Returns:
            None

        """
        # So, when solving with a correct area, there can be problems
        # That's because if the area's even slightly too large, it becomes infeasible
        if not self.area.fixed:
            self.area.value = self.area.value * 0.8

        initialise_inverted(self.hot_side, "deltaP")
        initialise_inverted(self.cold_side, "deltaP")

        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = self.hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1
        )

        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        flags2 = self.cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2
        )
        init_log.info_high("Initialization Step 1b (cold side) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.heat_transfer_equation.deactivate()
        if hasattr( self.cold_side.properties_out[0], "constraints"):
            self.cold_side.properties_out[0].constraints.deactivate()
        if hasattr( self.hot_side.properties_out[0], "constraints"):
            self.hot_side.properties_out[0].constraints.deactivate()

        # Get side 1 and side 2 heat units, and convert duty as needed
        s1_units = self.hot_side.heat.get_units()
        s2_units = self.cold_side.heat.get_units()

        # Check to see if heat duty is fixed
        # WE will assume that if the first point is fixed, it is fixed at all points
        if not self.cold_side.heat[self.flowsheet().time.first()].fixed:
            cs_fixed = False
            if duty is None:
                # Assume 1000 J/s and check for unitless properties
                if s1_units is None and s2_units is None:
                    # Backwards compatibility for unitless properties
                    s1_duty = -1000
                    s2_duty = 1000
                else:
                    s1_duty = pyunits.convert_value(
                        -1000, from_units=pyunits.W, to_units=s1_units
                    )
                    s2_duty = pyunits.convert_value(
                        1000, from_units=pyunits.W, to_units=s2_units
                    )
            else:
                # Duty provided with explicit units
                s1_duty = -pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s1_units
                )
                s2_duty = pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s2_units
                )

            self.cold_side.heat.fix(s2_duty)
            for i in self.hot_side.heat:
                self.hot_side.heat[i].value = s1_duty
        else:
            cs_fixed = True
            for i in self.hot_side.heat:
                self.hot_side.heat[i].set_value(self.cold_side.heat[i])
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        if not cs_fixed:
            self.cold_side.heat.unfix()
        if hasattr( self.cold_side.properties_out[0], "constraints"):
            self.cold_side.properties_out[0].constraints.activate()
        if hasattr( self.hot_side.properties_out[0], "constraints"):
            self.hot_side.properties_out[0].constraints.activate()
        self.heat_transfer_equation.activate()

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------

        # Release Inlet state
        self.hot_side.release_state(flags1, outlvl=outlvl)
        self.cold_side.release_state(flags2, outlvl=outlvl)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )


# Import Pyomo libraries
from ast import Expression
from cmath import pi
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from idaes.core.util.tables import create_stream_table_dataframe

from idaes.core.util.exceptions import ConfigurationError

# Set up logger
_log = idaeslog.getLogger(__name__)


# When using this file the name "Wind" is what is imported
@declare_process_block_class("Wind")
class WindData(UnitModelBlockData):
    """
    Zero order Link model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        #tmp_dict["defined_state"] = True  # inlet block is an inlet
       
        # Add outlet
        tmp_dict["defined_state"] = False  # outlet is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
       
        self.add_port(name="outlet", block=self.properties_out)

        # Add variable for efficiency
        self.efficiency = Var(
            initialize=1.0,
            doc="Efficiency of the turbine",
        )

        # Add variable for Windspeed
        self.windspeed = Var(self.flowsheet().config.time,
            initialize=1.0, units = pyunits.m/pyunits.s,
            doc="Speed of wind hitting the turbines",
        )

        # Add variable for Pressure
        self.pressure = Var(self.flowsheet().config.time,
            initialize=1.0, units = pyunits.Pa,
            doc="Air pressure",
        )

        # Add variable for Temperature
        self.air_temp = Var(self.flowsheet().config.time,
            initialize=1.0, units = pyunits.degK,
            doc="Air Temperature",
        )

        # Add variable for Humidity
        self.humidity = Var(self.flowsheet().config.time,
            initialize=1.0,
            doc="Air humidity",
        )

        # Add variable for rotor diameter
        self.rotor_diameter = Var(
            initialize=1.0, units = pyunits.m,
            doc="Rotor Diameter",
        )

        # Add variable for no. of turbines
        self.turbine_count = Var(
            initialize=1.0,
            doc="Number of turbines",
        )

        # Simple expressions
        self.radius = self.rotor_diameter / 2
        self.area = pi * self.radius**2
        self.Cp= 0.579  # Power coefficient, typically between 0.4 and 0.6 for wind turbines

        
        # Expression indexed by time
        @self.Expression(
            self.flowsheet().time
        )
        def density(b,t):
            return(
               (self.pressure[t]/(287.05 * self.air_temp[t])) * (1-0.378*(self.humidity[t]))
            )    


        # Add constraints
        @self.Constraint(
            self.flowsheet().time,
            doc="Power usage",
        )
        def eq_power_balance(b, t):
            return (
               0.5 * self.density[t] * self.area * self.windspeed[t]**3 * self.Cp * self.efficiency * self.turbine_count == b.properties_out[t].power
            )


    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):
        # Just propagate the power from inlet to outlet, good simple method of initialization
       pass

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe(
                {"Outlet": self.outlet}, time_point=time_point
            )
        except AttributeError:
            raise ConfigurationError(
                f"Unit model {self.name} does not have the standard Port "
                f"names (inlet and outlet). Please contact the unit model "
                f"developer to develop a unit specific stream table."
            )


# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.util.exceptions import ConfigurationError

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from pyomo.util.check_units import assert_units_consistent
from pyomo.environ import value

# Set up logger
_log = idaeslog.getLogger(__name__)


# When using this file the name "Link" is what is imported
@declare_process_block_class("Storage")
class StorageData(UnitModelBlockData):
    """
    Zero order Link model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
        self.properties_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )
        # Add outlet and waste block
        tmp_dict["defined_state"] = True  
        self.properties_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )


        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet", block=self.properties_in)
        self.add_port(name="outlet", block=self.properties_out)

        # Add variables:
        self.charging_efficiency = Var(
            self.flowsheet().config.time,
            initialize=1.0,
            doc="Charging Efficiency",
            )
        self.capacity = Var(self.flowsheet().config.time,
            initialize=1.0,
            units=pyunits.kWh,
            doc="Capacity of the storage",
            )
        self.initial_SOC = Var(self.flowsheet().config.time,
            initialize=0.4,
            doc="Initial State of Charge",
            )
        self.charging_power_in = Var(self.flowsheet().config.time,
            initialize=1.0,
            units=pyunits.W,
            doc="Power in",
            )
        self.charging_power_out = Var(self.flowsheet().config.time,
            initialize=1.0,
            units=pyunits.W,
            doc="Power out",
            )

        @self.Expression(
            self.flowsheet().time,
            doc="updated state of charge",
        )
        def updated_SOC(b,t):
            power_in = self.charging_power_in[t]
            power_out = self.charging_power_out[t]
            power_change = (power_in-power_out)/(pyunits.W *1000)
            capacity_without_unit = self.capacity[t]/pyunits.kWh
            if t == self.flowsheet().time.first():
                self.updated_SOC[t] = self.initial_SOC[t] + power_change/capacity_without_unit
            else:
                self.updated_SOC[t] = self.updated_SOC[t-1] + power_change/capacity_without_unit
          
            return self.updated_SOC[t]



        @self.Constraint(
            self.flowsheet().time,
            doc="Set output power for charging",
        )
        

        @self.Constraint(
            self.flowsheet().time,
            doc="Set output power for discharging",
        )
        def set_power_out_discharge(b,t): 
            return b.properties_out[t].power == self.charging_power_out[t]
       

        @self.Constraint(
            self.flowsheet().time,
            doc="Set output power for charging",
        )
        def set_power_charge(b,t):
            return self.charging_power_in[t] == self.properties_in[t].power

        
        # Add a constraint to ensure power_change is within a range
        @self.Constraint(
            self.flowsheet().time,
            doc="Ensure power_change is within a specified range",
        )
        def power_change_within_range(b, t):
            power_in = self.charging_power_in[t]
            power_out = self.charging_power_out[t]
            power_in_out = (power_in-power_out)/(pyunits.W *1000)
           # power_in_out = b.properties_out[t].power / (pyunits.W * 1000)
            remaining_power = power_in_out + self.initial_SOC[t] * self.capacity[t]
            return remaining_power <= self.capacity[t]
        
        @self.Constraint(
            self.flowsheet().time,
            doc="Ensure power_change is within capacity",
        )
        def power_change_above_zero(b, t):
            
            power_in = self.charging_power_in[t]
            power_out = self.charging_power_out[t]
            power_in_out = (power_in-power_out)/(pyunits.W *1000)
            #power_in_out = b.properties_out[t].power / (pyunits.W * 1000)
            remaining_power = power_in_out + self.initial_SOC[t] * self.capacity[t]
            return remaining_power >= 0
    
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):

        for i in blk.properties_in.index_set():
            if blk.initial_SOC[i].value == 0 and abs(blk.charging_power_out[i].value)> 0 and abs(blk.charging_power_in[i].value) == 0:
                raise ConfigurationError(
                    "Warning: There is no power to discharge in the battery."
                )
            if blk.charging_power_in[i].value > blk.capacity[i].value:
                pass
                #    raise ConfigurationError(
                #     "Warning: Battery capacity is exceeded!"
                # )       

    def _get_stream_table_contents(self, time_point=0):
        pass


import pytest

import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.power_generation.unit_models.waterpipe import WaterPipe
from property_packages.build_package import build_package
from idaes.core.util import DiagnosticsToolbox 

def test_waterpipe():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.water = build_package("helmholtz", ["water"], ["Liq"])
    m.fs.pipe = WaterPipe(property_package=m.fs.water)

    # Fix inlet conditions
    m.fs.pipe.inlet.flow_mol.fix(10)  
    m.fs.pipe.inlet.pressure.fix(101325)  
    # m.fs.pipe.inlet.temperature.fix(300)

    # Fix pipe parameters
    m.fs.pipe.length.fix(100)  # m
    m.fs.pipe.diameter.fix(0.5) 
    m.fs.pipe.number_of_pipes.fix(1)
    m.fs.pipe.elevation_change.fix(5) 
    m.fs.pipe.fcorrection_dp.fix(0.7)

    assert degrees_of_freedom(m.fs) == 0

    m.fs.pipe.initialize()
    solver = pyo.SolverFactory("ipopt")
    solver.options['max_iter']= 10000
    results = solver.solve(m, tee=False)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal


# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.exceptions import ConfigurationError
# Set up logger
_log = idaeslog.getLogger(__name__)



# When using this file the name "Transformer" is what is imported
@declare_process_block_class("Transformer")
class TransformerData(UnitModelBlockData):
    """
    Zero order Transformer model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Transformer unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
        self.properties_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )
        # Add outlet and waste block
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet", block=self.properties_in)
        self.add_port(name="outlet", block=self.properties_out)

        # # Add variable for turns_ratio
        # self.turns_ratio = Var(self.flowsheet().config.time,
        #     initialize=1.0,
        #     doc="Turns Ratio of the Transformer",
        # )
        # # Add variable for Load Resistance
        # self.resistance = Var(self.flowsheet().config.time,
        #     initialize=1.0,
        #     doc="Load Resistance of the Transformer",
        # )
        self.n_capacity = Var(
            self.flowsheet().config.time,
            initialize=1.0,
            units = pyunits.W,
            doc="N Capacity of the Transformer",
        )
        self.voltage = Var(self.flowsheet().config.time,
            initialize=1.0,
            units = pyunits.V,
            doc="Voltage Capacity of the Transformer",
        )
        self.efficiency = Var(self.flowsheet().config.time,
            initialize=1.0,
            units = pyunits.dimensionless,
            doc="Efficiency of the Transformer",
        )


        # Add constraints
        @self.Constraint(
            self.flowsheet().time,
            doc = "Power out",
        )
        def power_out(b, t):
            return (
                self.properties_out[t].power
                == self.properties_in[t].power * self.efficiency[t]
            )
        
        @self.Constraint(
            self.flowsheet().time,
            doc = "Capacity check",
        )
        def capacity_check(b,t):
            return abs(self.properties_in[t].power) <= self.n_capacity[t]

        # @self.Constraint(
        #     self.flowsheet().time,
        #     doc="Voltage out",
        # )
        # def voltage_out(b, t):
        #     return (
        #         b.properties_in[t].voltage * self.turns_ratio[t] == self.properties_out[t].voltage
        #     )
        # @self.Constraint(
        #     self.flowsheet().time,
        #     doc="Current out",
        # )
        # def current_out(b,t):
        #     return(
        #         self.properties_out[t].voltage / self.resistance[t] == self.properties_out[t].current
        #     )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):
        # Just propagate the power from inlet to outlet, good simple method of initialization
        for i in blk.properties_in.index_set():
            if not blk.properties_out[i].power.fixed:
                blk.properties_out[i].power = blk.properties_in[i].power.value
            
            if abs(blk.properties_in[i].power.value) > blk.n_capacity[i].value:
                 raise ConfigurationError(
                    "Danger: Input power exceeds transformer capacity. Please either increase capacity or lower input power."
                )


from typing import Any
from enum import Enum
from pyomo.core.base.units_container import units
from pyomo.environ import value as get_value
from ahuora_builder_types import FlowsheetSchema, UnitModelSchema, PropertiesSchema, PortsSchema
from ahuora_builder_types.flowsheet_schema import PropertyPackageType
from .methods.units_handler import idaes_specific_convert, attach_unit, get_attached_unit, get_attached_unit_str
from .methods.adapter_library import AdapterLibrary, UnitModelConstructor
from .methods import adapter_methods
from .build_state import get_state_vars
import json
from ahuora_builder.custom.energy.power_property_package import PowerParameterBlock



class Section:
    """
    Represents a section of a Python file
    Includes a name and a list of lines
    """
    def __init__(self, name: str, header: bool = True, new_line: bool = True, optional: bool = False) -> None:
        self._name: str = name
        self._lines: list[str] = []
        self._header: bool = header
        self._new_line: bool = new_line
        self._optional: bool = optional
    
    def extend(self, lines: list[str]) -> None:
        self._lines.extend(lines)
    
    def header(self) -> str:
        return f"### {self._name}"
    
    def lines(self) -> list[str]:
        return self._lines


class PythonFileGenerator:
    """
    Generate a Python file from the given model data
    """

    def __init__(self, schema: FlowsheetSchema) -> None:
        self._schema = schema
        
        # set some global constants
        self._model = "m"
        self._flowsheet = "fs"
        self._solver = "ipopt"
        # store the unit models, property packages, and ports
        self._property_packages: dict[int, dict[str, Any]] = {}  # {id: {name, vars: {vars}}}
        self._ports: dict[int, dict[str, Any]] = {}  # {id: {name, arc_id}}
        self._arcs: dict[int, dict[str, Any]] = {}  # {id: {name, source_id, destination_id}}

        # create the sections
        self._sections = {
            "imports": Section("Imports"),
            "property_package_imports": Section("Property Package Imports", header=False, new_line=False),
            "unit_model_imports": Section("Unit Model Imports", header=False, new_line=False),
            "utility methods": Section("Utility Methods"),
            "build_model": Section("Build Model"),
            "create_property_packages": Section("Create Property Packages", header=False),
            "create_unit_models": Section("Create Unit Models", header=False),
            "create_arcs": Section("Connect Unit Models", optional=True),
            "check_model": Section("Check Model Status"),
            "initialize": Section("Initialize Model"),
            "solve": Section("Solve"),
            "report": Section("Report"),
        }


    def sections(self) -> dict:
        return self._sections
    

    def setup_sections(self) -> None:
        """
        Set up the sections with the initial (constant) lines
        """
        self.extend("imports", [
            "from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination",
            "from pyomo.network import SequentialDecomposition, Port, Arc",
            "from pyomo.core.base.units_container import _PyomoUnit, units as pyomo_units",
            "from idaes.core import FlowsheetBlock",
            "from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom",
            "from idaes.core.util.tables import _get_state_from_port",
            "import idaes.logger as idaeslog",
        ])
        self.extend("property_package_imports", [
            "from property_packages.build_package import build_package",
        ])
        self.extend("utility methods", [
            "def units(item: str) -> _PyomoUnit:",
            "    ureg = pyomo_units._pint_registry",
            "    pint_unit = getattr(ureg, item)",
            "    return _PyomoUnit(pint_unit, ureg)",
        ])
        self.extend("build_model", [
            f"{self._model} = ConcreteModel()",
            f"{self._model}.{self._flowsheet} = FlowsheetBlock(dynamic=False)",
        ])
        self.extend("create_property_packages", [
            "# Set up property packages",
        ])
        self.extend("create_unit_models", [
            "# Create unit models",
        ])
        self.extend("check_model", [
            f"report_statistics({self._model})",
            f"print(\"Degrees of freedom:\", degrees_of_freedom({self._model}))",
        ])
        self.extend("solve", [
            f"opt = SolverFactory(\"{self._solver}\")",
            f"res = opt.solve({self._model}, tee=True)",
            f"assert_optimal_termination(res)",
        ])
    

    def extend(self, section_name: str, lines: list[str] | str) -> None:
        # add lines to a section
        if isinstance(lines, str):
            lines = [lines]
        self._sections[section_name].extend(lines)
    

    def add_section_excl(self, section_name: str, line: str) -> None:
        # add a line to a section if it is not already present
        if line not in self._sections[section_name].lines():
            self.extend(section_name, line)


    def resolve_import(self, obj: type | Enum) -> tuple[str, str]:
        # get the class name and import statement for a given class
        module_name = obj.__module__
        if isinstance(obj, Enum):
            import_name = obj.__class__.__name__  # name to be used in the import statement
            class_name = f"{obj.__class__.__name__}.{obj.name}"  # name to be used in the code
        else:
            import_name = obj.__name__
            class_name = import_name
        return class_name, f"from {module_name} import {import_name}"
    

    def get_name(self, name: str) -> str:
        # clean the name so that it can be used as a Python variable
        name = name.strip().replace("-", "_").replace(" ", "_")
        # remove any spaces and special characters
        name = "".join([char for char in name if char.isalnum() or char == "_"])
        # if the name is empty, use a default name
        if len(name) == 0:
            name = f"_unnamed_unit"
        # if the name starts with a number, add an underscore
        if name[0].isnumeric():
            name = "_" + name
        name = f"{self._model}.{self._flowsheet}.{name}"
        return name
    

    def create_property_packages(self) -> None:
        """
        Create property packages
        """
        for schema in self._schema.property_packages:
            self.create_property_package(schema)


    def create_property_package(self, schema: PropertyPackageType) -> None:
        name = self.get_name("PP_" + str(schema.id))
        compounds = schema.compounds
        phases = schema.phases
        type = schema.type
        self.extend("create_property_packages", [
        f"{name} = build_package(",
        f"    \"{type}\",",
        f"    {json.dumps(compounds)},",
        ")",
        ])
        self._property_packages[schema.id] = { "name": name, "vars": get_state_vars(schema) }


    def get_property_package(self, id: int) -> dict:
        """Get the name of a property package by ID"""
        if id == -1 and id not in self._property_packages:
            # add a default Helmholtz property package (for testing purposes)
            self.create_property_package(PropertyPackageType(id=id, type="helmholtz", compounds=["h2o"], phases=["Liq"]))
        return self._property_packages[id]
    
    def get_power_property_package(self, id: str):
        power_package = PowerParameterBlock
        return self._property_packages[-1]


    def get_property_package_at_port(self, model_schema: UnitModelSchema, port: str) -> dict:
        """Get the property package that is used at a port"""
        # generally, the unitop has one property package that is used for all ports
        if model_schema.args.get("property_package") is not None:
            return self.get_property_package(model_schema.args["property_package"])
        # this is a special case (hard-coded for now) for heat exchangers, which have two property packages
        # TODO: make this dynamic once we have parent stream inheritance
        package_arg = port.removesuffix("_inlet").removesuffix("_outlet")  # eg. "hot_side_inlet" -> "hot_side"
        return self.get_property_package(model_schema.args[package_arg]["property_package"])
    

    def serialise_dict(self, d: dict, indent: bool = False, indent_level: int = 1, nested_indent: bool = True) -> str:
        if len(d) == 0:
            return "{}"
        result = "{"
        for k, v in d.items():
            if indent:
                result += "\n" + "    " * indent_level
            adj_k = f"\"{k}\"" if isinstance(k, str) else k
            adj_v = f"\"{v}\"" if isinstance(v, str) and not v.startswith(f"{self._model}.{self._flowsheet}.") else v
            if isinstance(v, dict):
                # allow nested dictionaries
                adj_v = self.serialise_dict(v, indent=indent and nested_indent, indent_level=indent_level + 1)
            result += f"{adj_k}: {adj_v},"
        result = result[:-1] + ("\n" + "    " * (indent_level - 1)) * indent + "}"
        return result
    

    def serialise_list(self, l: list) -> str:
        if len(l) == 0:
            return "[]"
        result = "["
        for v in l:
            adj_v = f"\"{v}\"" if isinstance(v, str) and not v.startswith(f"{self._model}.{self._flowsheet}.") else v
            result += f"{adj_v},"
        result = result[:-1] + "]"
        return result
    

    def setup_args(self, args: dict, arg_parsers: dict) -> dict:
        """Setup the arguments for a unit model"""
        result: dict[str, Any] = {}
        print("args: " + str(args))
        for arg_name, method in arg_parsers.items():
            def match_method() -> Any:
                match method.__class__:
                    case adapter_methods.Constant:
                        # constant, defined in the method
                        constant = method.run(None,None)
                        # constant can be a function, in which case we need to resolve the import
                        if callable(constant) or isinstance(constant, Enum):
                            constant_class_name, constant_import = self.resolve_import(constant)
                            self.add_section_excl("imports", constant_import)
                            return constant_class_name
                        return constant
                    case adapter_methods.Value:
                        # value, keep as is
                        return args.get(arg_name, None)
                    case adapter_methods.PropertyPackage:
                        # property package
                        property_package_id = args["property_package"]
                        print(property_package_id)
                        return self.get_property_package(property_package_id)["name"]
                    case adapter_methods.PowerPropertyPackage:
                        #power property package
                        return "m.fs.power_property_package"
                    
                    case adapter_methods.Dictionary:
                        # another dictionary of arg parsers, recursively setup the args
                        return self.setup_args(args[arg_name], method._schema)
                    case _:
                        raise Exception(f"Method {method} not supported")
            result[arg_name] = match_method()
        return result


    def write_args(self, args: dict) -> str:
        args_str = ""
        if len(args) == 0:
            return args_str
        args_str += "\n"
        for key, value in args.items():
            args_str += f"    {key}="
            if isinstance(value, dict):
                args_str += self.serialise_dict(value)
            else:
                args_str += f"{value}"
            args_str += ",\n"
        args_str = args_str[:-2] + "\n"  # remove the last comma
        return args_str
    

    def create_unit_models(self) -> None:
        """
        Create the unit models
        """
        for unit_model in self._schema.unit_models:
            # add to imports
            adapter: Adapter = AdapterLibrary[unit_model.type]
            class_name, class_import = self.resolve_import(adapter.model_constructor)
            self.add_section_excl("unit_model_imports", class_import)
            # setup args
            args = self.setup_args(unit_model.args, adapter.arg_parsers)
            args_str = self.write_args(args)
            print("args_str: " + args_str)

            # create the unit model
            name = self.get_name(unit_model.name)
            self.extend("create_unit_models", f"\n# {unit_model.name}")  # comment
            self.extend("create_unit_models", f"{name} = {class_name}({args_str})")  # constructor
            self.extend("create_unit_models", self.fix_properties(name, unit_model.properties))  # fix properties
            for port_name, port_data in unit_model.ports.items():
                # save the port
                global_name = f"{name}.{port_name}"
                self._ports[port_data.id] = { "name": global_name, "arc": None }
                # available_vars = self.get_property_package_at_port(unit_model, port_name)["vars"]
                # fix the properties of the port
                self.extend("create_unit_models", self.fix_state_block(global_name, port_data.properties))
            self.extend("report", f"{name}.report()")
    

    def fix_properties(self, prefix: str, properties_schema: PropertiesSchema) -> list[str]:
        lines = []
        for key, property_info in properties_schema.items():
            for property_value in property_info.data:
                if property_value.value is None:
                    continue
                if property_value.discrete_indexes is not None:
                    indexes_tuple = tuple(property_value.discrete_indexes)
                    indexes_string = f"[{indexes_tuple}]" if len(property_value.discrete_indexes) > 0 else ""
                else:
                    indexes_string = ""
                val = get_value(property_value.value)
                unit = property_info.unit
                # TODO: Handle dynamic indexes etc.
                lines.append(f"{prefix}.{key}{indexes_string}.fix({val} * units(\"{unit}\"))")
                    
        return lines

    
    def fix_state_block(self, prefix: str, properties: PropertiesSchema) -> list[str]:
        """
        Fix the properties of a unit model
        """
        lines = []
        lines.append(f"sb = _get_state_from_port({prefix}, 0)")
        for key, property_info in properties.items():
            for property_value in property_info.data:
                if property_value.value is None:
                    continue
                if property_value.discrete_indexes is None:
                    indexes_str = ""
                else:
                    # We aren't worrying about time yet, but we will need to do in the future.
                    indexes_tuple = tuple(property_value.discrete_indexes)
                    indexes_str = f"[{indexes_tuple}]" if len(property_value.discrete_indexes) > 0 else ""
                val = get_value(property_value.value)
                unit = property_info.unit
                lines.append(f"sb.constrain_component(sb.{key}{indexes_str}, {val} * units(\"{unit}\"))")
                    
        if len(lines) == 1:
            return []
        return lines
    

    def create_arcs(self) -> None:
        """
        Create the arcs
        """
        if len(self._schema.arcs) == 0:
            return
        for i, arc in enumerate(self._schema.arcs):
            source = self._ports[arc.source]
            destination = self._ports[arc.destination]
            name = f"{self._model}.{self._flowsheet}.arc_{i + 1}"
            self.extend("create_arcs", f"{name} = Arc(source={source['name']}, destination={destination['name']})")
            self._arcs[i] = { "name": name, "source": arc.source, "destination": arc.destination }
            source["arc"] = i
            destination["arc"] = i
        # add to initialization: expand the arcs
        self.extend("initialize", f"TransformationFactory(\"network.expand_arcs\").apply_to({self._model})")
    

    def initialize(self) -> None:
        """
        Initialize the model
        """
        def is_connected(ports: PortsSchema) -> bool:
            for _, port_data in ports.items():
                port = self._ports[port_data.id]
                if port["arc"] is not None:
                    return True
            return False
        # initialize everything that is not connected
        for unit_model in self._schema.unit_models:
            if not is_connected(unit_model.ports):
                name = self.get_name(unit_model.name)
                self.extend("initialize", f"{name}.initialize(outlvl=idaeslog.INFO)")
        if len(self._schema.arcs) == 0:
            return
        # setup sequential decomposition
        self.extend("utility methods", [
            "\ndef init_unit(unit: Block) -> None:",
            "    unit.initialize(outlvl=idaeslog.INFO)"
        ])
        self.extend("initialize", "seq = SequentialDecomposition()")
        # set tear guesses
        tear_set = []

        # Need to rewrite the logic for dealing with tear sets from recycle

        self.extend("initialize", f"seq.set_tear_set({self.serialise_list(tear_set)})")
        self.extend("initialize", f"seq.run({self._model}, init_unit)")
                

def generate_python_code(model_data: FlowsheetSchema) -> str:
    """
    Generate a Python file from the given model data
    """
    generator = PythonFileGenerator(model_data)
    generator.setup_sections()
    generator.create_property_packages()
    generator.create_unit_models()
    generator.create_arcs()
    generator.initialize()
    sections = generator.sections()

    result = ""
    for key, section in sections.items():
        if section._optional and len(section.lines()) == 0:
            # skip empty sections
            continue
        # add extra newline characters between sections
        if section._new_line and key != list(sections.keys())[0]:
            if section._header:
                result += "\n\n"
            else:
                result += "\n"
        # add section header
        if section._header:
            result += section.header()
            if section._new_line:
                result += "\n"
        # write each line to the result string
        # separated by a newline character
        result += "\n".join(section.lines()) + "\n"
    return result
    


from collections import defaultdict, deque
import re
from pyomo.network import SequentialDecomposition, Port
from pyomo.environ import (
    ConcreteModel,
    TransformationFactory,
    SolverFactory,
    Block,
    Expression,
    Constraint,
    Objective,
    minimize,
    assert_optimal_termination,
    ScalarVar,
    units as pyunits,
    TerminationCondition
)
from pyomo.core.base.constraint import ScalarConstraint, IndexedConstraint
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
import idaes.logger as idaeslog
from idaes.core.util import DiagnosticsToolbox
from ahuora_builder_types import FlowsheetSchema
from ahuora_builder_types.flowsheet_schema import SolvedFlowsheetSchema
from .property_package_manager import PropertyPackageManager
from .port_manager import PortManager
from .arc_manager import ArcManager
from .tear_manager import TearManager
from .unit_model_manager import UnitModelManager
from .methods.adapter_library import AdapterLibrary
from .methods.adapter import (
    serialize_properties_map,
    deactivate_fixed_guesses,
    deactivate_component,
    deactivate_components,
    add_corresponding_constraint,
)
from .methods.expression_parsing import parse_expression, ExpressionParsingError
from .timing import start_timing
from .methods.units_handler import get_value, get_attached_unit, check_units_equivalent
from .properties_manager import PropertiesManager
from .custom.energy.power_property_package import PowerParameterBlock
from .custom.energy.ac_property_package import acParameterBlock
from .custom.energy.transformer_property_package import transformerParameterBlock 
from pyomo.core.base.units_container import units, _PyomoUnit
from idaes.core.util.model_serializer import StoreSpec, from_json, to_json
from .diagnostics.infeasibilities import print_infeasibilities

# CSTR Imports from example flowsheet. 
# To be used as placeholders until bespoke functions can be developed.
from .custom import hda_reaction as reaction_props
from .custom.hda_ideal_VLE import HDAParameterBlock
from ahuora_builder.properties_manager import PropertyComponent
from property_packages.build_package import build_package

# from amplpy import modules
# Import required to allow the library to set the PATH and allow conopt to be found.

def build_flowsheet(dynamic=False,time_set=[0]):
    """
    Builds a flowsheet block
    """
    # create the model and the flowsheet block
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=dynamic, time_set=time_set, time_units=units.s)
    model.fs.guess_vars = []
    model.fs.controlled_vars = Block()
    # Always add the power property package (as there's only one and there's no different property package types)
    # TODO: should this be on-demand?
    model.fs.power_pp = PowerParameterBlock()
    model.fs.ac_pp = acParameterBlock()
    model.fs.tr_pp = transformerParameterBlock()
    # properties map: { id: pyomo variable or expression }
    # used to map symbols for a sympy expression
    model.fs.properties_map = PropertiesManager()
    # list of component constraints to add later
    model.fs.constraint_exprs = []

    # Placholder property packages for CSTR reactor.
    model.fs.BTHM_params = HDAParameterBlock()
    #Hard-coded peng-robinson package for reactor:
    model.fs.peng_robinson = build_package("peng-robinson",["benzene", "toluene", "hydrogen", "methane"], ["Liq","Vap"])
    
    # Reaction package for the HDA reaction
    model.fs.reaction_params = reaction_props.HDAReactionParameterBlock(
       property_package=model.fs.peng_robinson
    )

    return model


class FlowsheetManager:
    """
    Manages the flowsheet, including the property packages, unit models, ports, arcs, and tears
    Includes methods to load, initialise, and solve the flowsheet
    """

    def __init__(self, schema: FlowsheetSchema) -> None:
        """
        Stores all relevant information about the flowsheet, without actually loading it
        """
        self.timing = start_timing()
        self.timing.add_timing("initialise_flowsheet_manager")

        self.model = build_flowsheet(dynamic=schema.dynamic, time_set=schema.time_set)

        self.schema = schema
        # Add property packages first, so that unit models can use them
        self.property_packages = PropertyPackageManager(self)
        # Add the port manager, so the unit models can register their ports
        self.ports = PortManager()
        # Add unit models
        self.unit_models = UnitModelManager(self)
        # Add arcs to connect the unit models together
        self.arcs = ArcManager(self)
        # set certain arcs as tears
        self.tears = TearManager(self)

    def load(self) -> None:
        """
        Parses the schema and loads the model
        """
        self.timing.step_into("load_flowsheet")
        # Load property packages first, so that unit models can use them
        self.timing.add_timing("load_property_packages")
        self.property_packages.load()
        self.timing.step_into("load_unit_models")
        self.unit_models.load()
        self.timing.step_out()
        # no need to load ports seperately, they are loaded by the unit models
        # Load arcs to connect the unit models together
        self.arcs.load()
        # load any expressions
        self.load_specs()
        # if dynamics, apply finite difference transformaiton
        if self.schema.dynamic:
            print("performing finite difference with", len(self.model.fs.time), "time steps")
            TransformationFactory("dae.finite_difference").apply_to(
                self.model.fs, 
                nfe=len(self.model.fs.time)-1, # Number of finite elements to use for discretization. We aren't adding any extra steps as our constraints dont work for that.
                wrt=self.model.fs.time, 
                scheme="BACKWARD"
            )
            

        self.timing.step_out()

    def load_specs(self) -> None:
        """
        Loads expressions from the schema
        """
        fs = self.model.fs
        specs = self.schema.expressions or []

        ## Build dependency tree and sort specs before loading.
        dependencies, result_ids = self.build_spec_dependency_tree(specs)
        sorted_result_ids = self.topological_sort(dependencies, result_ids)

        # load the specs (expressions within specifications tab)
        for result_id in sorted_result_ids:
            spec_config = next(spec for spec in specs if spec["id"] == result_id)
            expression_str = spec_config["expression"]
            try:
                component_name = f"{spec_config['name']}_{spec_config['id']}"
                def expression_rule(blk, time_index):
                    return parse_expression(expression_str, fs,time_index)
                
                component = Expression(fs.time, rule=expression_rule)
                fs.add_component(component_name, component)
                fs.properties_map.add(
                    spec_config["id"], component, component.name, unknown_units=True
                )
            except ExpressionParsingError as e:
                raise ExpressionParsingError(f"{e} when parsing expression '{expression_str}' for {component_name}: ")

        # load constraints (expressions for specific property infos)
        # can only handle equality constraints for now 
        for component, expr_str, id in fs.constraint_exprs:
            # get the time index
            #for time_index in fs.time:
            if component.index_set().dimen != 0 and component.index_set() != fs.time:
                raise ExpressionParsingError(f"Cannot add constraint for {component}: only time-indexed components are supported.") 
            try:
                def constraint_rule(blk, time_index):
                    expression = parse_expression(expr_str, fs,time_index)
                    # make sure the units of the expression are the same as the component
                    u1, u2 = get_attached_unit(component), get_attached_unit(expression)
                    if not check_units_equivalent(u1, u2):
                        raise ValueError(
                            f"Failed to add constraint for {component}: units do not match (expected {u1}, got {u2})"
                        )
                    return pyunits.convert(component[time_index], to_units=u2) == expression
                c = Constraint(component.index_set(), rule= constraint_rule)
                name = f"equality_constraint_{id}"
                fs.add_component(name, c)
                add_corresponding_constraint(fs, c, id)
            except ExpressionParsingError as e:
                raise ExpressionParsingError(f"Failed to parse constraint expression '{expr_str}'  for {component}: {expr_str}, error: {e}")


    def build_spec_dependency_tree(self, specs) -> tuple:
        """
        Builds dependency tree for expressions based on the references in their respective expressions.
        """
        dependencies = defaultdict(
            set
        )  # Maps an expression's result_id to a set of result_ids it depends on
        result_ids = set()  # A set to track all result_ids

        # Get a list of all result_id's.
        for spec in specs:
            result_id = spec["id"]
            result_ids.add(result_id)

        for spec in specs:
            result_id = spec["id"]
            expression = spec["expression"]

            # Find the result_ids that this expression depends on
            dependent_expressions = self.get_dependent_expressions(
                expression, result_ids
            )

            if dependent_expressions:
                # If the expression depends on another result_id, add dependency
                for id in dependent_expressions:
                    dependencies[id].add(result_id)

        return dependencies, result_ids

    def get_dependent_expressions(self, expression: str, all_result_ids: set) -> list:
        """
        Gets all result_ids referenced in the expression.
        """
        # match result_ids starting with 'id_' followed by numbers
        ids = re.findall(r"\b(id_\d+)\b", expression)

        # Filter the referenced_ids to only include those that are in all_result_ids
        valid_referenced_ids = []
        for id in ids:
            # get the numeric part after "id_" and check if it's in the all_result_ids
            numeric_id = int(id[3:])
            if numeric_id in all_result_ids:
                valid_referenced_ids.append(numeric_id)

        return valid_referenced_ids

    def topological_sort(self, dependencies: dict, result_ids: set) -> list:
        """
        Performs topological sorting on the specification dependency tree.
        """
        # Track teh in-degree count for all expressions (edges coming into it)
        in_degree = defaultdict(int)

        # Count dependencies for each expression
        for result_id in result_ids:
            for dep in dependencies[result_id]:
                in_degree[dep] += 1

        # Initialise the queue with result_ids that have no dependencies (in-degree 0)
        dequeue = deque(
            [result_id for result_id in result_ids if in_degree[result_id] == 0]
        )

        sorted_result_ids = []

        while dequeue:
            result_id = dequeue.popleft()
            sorted_result_ids.append(result_id)

            # loop thtough and decrement each dependent expression's in_degree, and append it to the deque if it has an in-degree of 0.
            for dependent_result_id in dependencies[result_id]:
                in_degree[dependent_result_id] -= 1
                if in_degree[dependent_result_id] == 0:
                    dequeue.append(dependent_result_id)

        # If there are any result_ids left with non-zero in-degree, a cycle exists so error.
        if len(sorted_result_ids) != len(result_ids):
            raise ValueError(
                "Cycle detected in the dependency graph. Check an expression does not reference itself!"
            )

        return sorted_result_ids

    def initialise(self) -> None:
        """
        Expands the arcs and initialises the model
        """
        # check if initialisation is disabled for this scenario
        if getattr(self.schema, "disable_initialization", False):
            # if disable initialisation is set to True, then we don't need to initialise the model
            print("Initialisation is disabled for this scenario.")

        # We need to "expand_arcs" to make them a bidirection link that actually imposes constraints on the model.
        TransformationFactory("network.expand_arcs").apply_to(self.model)

        self.timing.step_into("initialise_model")

        # load tear guesses (including var/constraint unfixing & deactivation and/or equality constraint deactivation)
        self.tears.load()

        tears = self.tears._tears

        def init_unit(unit):
            if getattr(self.schema, "disable_initialization", False):
                return

            print(f"Initializing unit {unit}")
            self.timing.add_timing(f"init_{unit.name}")
            #unit.display()
            unit.initialize(outlvl=idaeslog.INFO)
            #unit.report()

        self.timing.add_timing("setup_sequential_decomposition")
        # Use SequentialDecomposition to initialise the model
        seq = SequentialDecomposition(
            #run_first_pass=True,
            iterLim=1,
        )
        seq.set_tear_set(tears)
        # use create_graph to get the order of sequential decomposition, and also to
        # find any units that are not connected to the sequential decomposition
        G = seq.create_graph(self.model)
        order = seq.calculation_order(G)
        seq_blocks = []
        for o in order:
            seq_blocks.append(o[0])
        print("Order of initialisation:", [blk.name for blk in seq_blocks])
        # set all the tear guesses before running the decomposition
        for arc in tears:
            port = arc.destination
            # guesses used are initial values for each var
            guesses = {}
            guesses = {key: get_value(var) for key, var in port.vars.items()}
            print(f"Guess for {port}: {guesses}")

        self.timing.step_into("run_sequential_decomposition")

        # sequential decomposition completes when all vars across port
        # equalities are within tol of each other
        seq.options["tol"] = 1e-2
        # seq.options["solve_tears"] = False
        res = seq.run(self.model, init_unit)

        self.timing.step_out()
        self.timing.add_timing("initialise_disconnected_units")
        # Initialise any unit model that is not connected to the sequential decomposition
        for blk in self.model.fs.component_data_objects(
            Block, descend_into=False, active=True
        ):
            ports = list(blk.component_objects(Port, descend_into=False))
            if len(ports) == 0:
                continue  # if the block has no ports, then it is not a unit model
            if blk in seq_blocks:
                continue  # already initialised by sequential decomposition
            init_unit(blk)

        # unfix guess vars
        deactivate_fixed_guesses(self.model.fs.guess_vars)

        self.timing.step_out()

    def serialise(self) -> SolvedFlowsheetSchema:
        self.timing.add_timing("serialise_model")

        initial_values = {}
        for unit_model_id, unit_model in self.unit_models._unit_models.items():
            initial_values[str(unit_model_id)] = to_json(unit_model, return_dict=True, wts=StoreSpec.value())

        solved_flowsheet = SolvedFlowsheetSchema(
            id=self.schema.id,
            properties=serialize_properties_map(self.model.fs),
            initial_values=initial_values
        )

        return solved_flowsheet

    def report_statistics(self) -> None:
        """
        Reports statistics about the model
        """
        report_statistics(self.model)
        # I think the diagnostics toolbox is taking too long to run, so commenting it out for now.
        # dt = DiagnosticsToolbox(self.model)
        # dt.report_structural_issues()
        # dt.display_overconstrained_set()
        # dt.display_underconstrained_set()
    
    def diagnose_problems(self) -> None:
        print("=== DIAGNOSTICS ===")
        report_statistics(self.model)
        dt = DiagnosticsToolbox(self.model)
        dt.report_structural_issues()
        dt.display_overconstrained_set()
        dt.display_underconstrained_set()
        #dt.display_components_with_inconsistent_units()
        try:
            dt.compute_infeasibility_explanation()
        except Exception as e:
            print(f"{e}") # error is probably because it is feasible
        dt.report_numerical_issues()
        dt.display_near_parallel_constraints()
        dt.display_variables_at_or_outside_bounds()
        print("=== END DIAGNOSTICS ===")

    def degrees_of_freedom(self) -> int:
        """
        Returns the degrees of freedom of the model
        """
        return int(degrees_of_freedom(self.model))

    def check_model_valid(self) -> None:
        """
        Checks if the model is valid by checking the
        degrees of freedom. Will raise an exception if
        the model is not valid.
        """
        self.timing.add_timing("check_model_valid")
        degrees_of_freedom = self.degrees_of_freedom()
        if degrees_of_freedom != 0:
            #self.model.display()  # prints the vars/constraints for debugging
            raise Exception(
                f"Degrees of freedom is not 0. Degrees of freedom: {degrees_of_freedom}"
            )

    def solve(self) -> None:
        """
        Solves the model
        """
        self.timing.add_timing("solve_model")
        print("=== Starting Solve ===")

        opt = SolverFactory(self.schema.solver_option)
        # opt.options["max_iter"] = 5000

        if self.schema.solver_option != "conopt":
            opt.options["max_iter"] = 1000
        try:
            res = opt.solve(self.model, tee=True)
            if res.solver.termination_condition != TerminationCondition.optimal:
                print_infeasibilities(self.model.fs.properties_map)
            assert_optimal_termination(res)
        except ValueError as e:
            if str(e).startswith("No variables appear"):
                # https://github.com/Pyomo/pyomo/pull/3445
                pass
            else:
                raise e

    def optimize(self) -> None:
        if self.schema.optimizations is None or self.schema.optimizations == []:
            return

        self.timing.add_timing("optimize_model")
        print("=== Starting Optimization ===")

        # ipopt doesn't support multiple objectives, so we need to create
        # a single objective expression.
        # this is done by summing all objectives in the model, adding
        # or subtracting based on the sense (minimize or maximize)
        objective_expr = 0
        for schema in self.schema.optimizations:
            # get the expression component to optimize
            # TODO: This is assuming optimisation is run on a steady-state simulation. We need to change how this works to handle dynamics.
            # For now, just hardcoding time_index=0
            objective_component = self.model.fs.properties_map.get_component(schema.objective)
            # the objective component should be a scalar in non-dynamic models
            # in a dynamic model it'll be indexed across all time steps
            # We sum up all time steps so each time step is weighted equally.
            objective = sum(objective_component.values())
            # add or subtract the objective based on the sense
            sense = schema.sense
            if sense == "minimize":
                objective_expr += objective
            else:
                objective_expr -= objective



            # unfix relevant vars (add to degrees of freedom)
            for dof_info in schema.unfixed_variables:
                id = dof_info.id # Id of the propertyValue for this degree of freedom
                var: PropertyComponent = self.model.fs.properties_map.get(id)
        

                # TODO: may need to handle deactivating constraints,
                # for expressions that are constrained (instead of state vars)
                c = var.corresponding_constraint
                if c is not None:
                    # TODO: better typing for constraints
                    if isinstance(c, ScalarConstraint) or isinstance(c, IndexedConstraint):
                        c.deactivate()
                    else:
                        deactivate_components(c)
                else:
                    for i in var.component.values():
                        if isinstance(i, ScalarVar):
                            i.unfix()
                        # Because if not, it is ExpressionData, meaning it is already an expression and doesn't need to be unfixed. (we've already checked if there is a constraint for it above too.)

                
                # TODO: set attributes for upper and lower bounds of property infos. i.e. use propertyinfo id.
                # Var is either a Variable or Expression
                # set the minimum or maximum bounds for this variable if they are enabled
                #self.model.upper_bound_12 =   Constraint(expr= var <= upper_bound_value )

                upper_bound = dof_info.upper_bound
                lower_bound = dof_info.lower_bound

                c = var.component

                if upper_bound is not None:
                    def upper_bound_rule(model,index):
                        return c[index] <= upper_bound
                    upper_bound_constraint = Constraint(c.index_set(),rule=upper_bound_rule)
                    setattr(self.model,"upper_bound_" + str(id), upper_bound_constraint)

                if lower_bound is not None:
                    def lower_bound_rule(model,index):
                        return c[index] >= lower_bound
                    lower_bound_constraint = Constraint(c.index_set(),rule=lower_bound_rule)
                    setattr(self.model,"lower_bound_" + str(id), lower_bound_constraint)


        # add the objective to the model
        self.model.objective = Objective(expr=objective_expr, sense=minimize)

        # solve the model with the objective
        opt = SolverFactory(self.schema.solver_option)

        if self.schema.solver_option != "conopt":
            opt.options["max_iter"] = 1000

        try:
            res = opt.solve(self.model, tee=True)
            assert_optimal_termination(res)
        except ValueError as e:
            if str(e).startswith("No variables appear"):
                # https://github.com/Pyomo/pyomo/pull/3445
                pass
            else:
                raise e



import os
import json
import io

from typing import Any

from ahuora_builder_types import FlowsheetSchema
from ahuora_builder_types.payloads import BuildStateRequestSchema
from ahuora_builder.flowsheet_manager import FlowsheetManager
from ahuora_builder.build_state import solve_state_block
from contextlib import redirect_stdout


# Set this to True to regenerate expected output files, if needed (e.g output file format changes.)
# Make sure to review the commit to ensure that only intended changes are made to expected output files.
GENERATE_TESTS = False

# Get current location (so that we can retrieve pump.json)
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def try_solve(input_file: str, expected_output_file: str, state_block=False) -> None:
    """Tests that the model can be solved and that the answer returned is correct"""

    with open(os.path.join(__location__, input_file), 'r') as file:

        data = json.load(file)

        if state_block:
            build_state_schema = BuildStateRequestSchema.model_validate(data)
            result = solve_state_block(build_state_schema)
            result = [solved_value.model_dump() for solved_value in result]
        else:
            flowsheet_schema = FlowsheetSchema.model_validate(data)
            flowsheet = FlowsheetManager(flowsheet_schema)
            flowsheet.load()
            flowsheet.initialise()
            assert flowsheet.degrees_of_freedom() == 0, "Degrees of freedom is not 0: " + str(flowsheet.degrees_of_freedom())
            flowsheet.report_statistics()
            flowsheet.solve()
            flowsheet.optimize()

            result = flowsheet.serialise()
            result = result.model_dump()

            # ignore initial values
            del result["initial_values"]

        # # write result to a file for debugging - or to regenerate expected files
        if GENERATE_TESTS:
            with open(os.path.join(__location__, expected_output_file), 'w') as f:
                f.write(json.dumps(result, indent=4))

        print(result)

        # Assert - Check if all the parameters are correct
        with open(os.path.join(__location__, expected_output_file), 'r') as file:
            # Load the result JSON file
            expected_result = json.load(file)

            # Check if the result is equal to the expected_result
            try:
                if state_block:
                 values_approximately_equal(result, expected_result, 1e-1)
                else:
                    values_approximately_equal(result, expected_result, 1e-1)
            except Exception as e:
                print(e)
                raise AssertionError(f"Did not solve the model correctly")


def values_approximately_equal(val1: Any, val2: Any, tolerance: float = 1e-3) -> bool:
    """
    Checks if two values are approximately equal
    """
    if isinstance(val1, (float, int)) and isinstance(val2, (float, int)):
        if not abs(val1 - val2) < tolerance:
            raise AssertionError(f"Value: {val1} is not equal to expected value: {val2} within tolerance {tolerance}")
    elif isinstance(val1, list) and isinstance(val2, list):
        if len(val1) != len(val2):
            raise AssertionError(f"Lists have different lengths: {len(val1)} and {len(val2)}")
        return all(values_approximately_equal(v1, v2, tolerance) for v1, v2 in zip(val1, val2))
    elif isinstance(val1, dict) and isinstance(val2, dict):
        if set(val1.keys()) != set(val2.keys()):
            raise AssertionError(f"Keys: {val1.keys()} are not equal to expected keys: {val2.keys()}")
        return all(values_approximately_equal(val1[key], val2[key], tolerance) for key in val1)
    else:
        if val1 != val2:
            raise AssertionError(f"Value: {val1} is not equal to expected value: {val2}")
    return True


def test_generate_is_false():
    assert GENERATE_TESTS is False, "test_generate_python_file.py:  GENERATE_TESTS should be False to avoid overriding test_solved.json files during tests."


def test_solve_pump_and_heater() -> None:
    # Tests that the pump and heater model can be solved 
    try_solve('configurations/pump.json', 'configurations/pump_solved.json')


def test_solve_recycle() -> None:
    # Tests that the recycle model can be solved (Recycle connected to itself)
    try_solve('configurations/recycle.json', 'configurations/recycle_solved.json')


def test_solve_compressor() -> None:
    # Tests that the compressor model can be solved
    try_solve('configurations/compressor.json', 'configurations/compressor_solved.json')


def test_solve_mixer() -> None:
    # Tests that the mixer model can be solved
    try_solve('configurations/mixer.json', 'configurations/mixer_solved.json')


def test_solve_bt_pr() -> None:
    # Peng-Robinson test with Benzene and Toluene
    try_solve('configurations/BT_PR.json', 'configurations/BT_PR_solved.json')


def test_solve_heat_exchanger() -> None:
    """Tests that the BT_PR model can be solved"""
    try_solve('configurations/heat_exchanger.json','configurations/heat_exchanger_solved.json')


def test_solve_heat_pump() -> None:
    """Tests that the turbine model can be solved"""
    # TODO: fix this test case - the values specified are not very good
    # and although the model solves, it doesn't really show anything useful
    try_solve('configurations/heat_pump.json','configurations/heat_pump_solved.json')


def test_solve_unit_conversions() -> None:
    """
    Checks that unit conversions work as expected

    To do this, we solve the pump model and check that everything works as expected
    """
    # the output file should stay the same since we are just changing the units
    try_solve('configurations/pump_unit_conversions.json', 'configurations/pump_solved.json')


def test_solve_expressions() -> None:
    """
    Checks that expressions work as expected
    """
    try_solve('configurations/expressions.json', 'configurations/expressions_solved.json')


def test_solve_vapor_frac_target():
    # compressor_test_with_vapor_fraction - targets an outlet vapor fraction of 1
    try_solve('configurations/vapor_frac_target.json', 'configurations/vapor_frac_target_solved.json')


def test_solve_mass_flow_tear():
    # recycle with mass flow constraint
    try_solve('configurations/mass_flow_tear.json', 'configurations/mass_flow_tear_solved.json')


def test_solve_propane_recycle():
    try_solve('configurations/propane_recycle.json', 'configurations/propane_recycle_solved.json')


def test_solve_propane_heat_pump():
    try_solve('configurations/propane_heat_pump.json', 'configurations/propane_heat_pump_solved.json')


def test_solve_control():
    # inlet enthalpy is controlled by the set point of the outlet enthalpy
    try_solve('configurations/control.json', 'configurations/control_solved.json')
    
def test_solve_machine_learning():
    try_solve('configurations/machine_learning.json', 'configurations/machine_learning_solved.json')


def test_solve_sb_vapor_frac() -> None:
    # Tests that a state block with vapor_frac can be solved
    try_solve('configurations/sb_vapor_frac.json', 'configurations/sb_vapor_frac_solved.json', state_block=True)


def test_solve_compound_separator() -> None:
    # Tests that a compound separator model can be solved
    try_solve('configurations/compound_separator.json', 'configurations/compound_separator_solved.json')


def test_solve_constraints() -> None:
    # Tests that a model with constraints can be solved
    # Tests for an equality constraint using an expression
    try_solve('configurations/constraints.json', 'configurations/constraints_solved.json')


def test_solve_link() -> None:
    # Tests that a bus model can be solved
    try_solve('configurations/bus.json', 'configurations/bus_solved.json')


def test_solve_optimization() -> None:
    # Tests optimizing an expression
    try_solve('configurations/optimization.json', 'configurations/optimization_solved.json')

def test_solve_solar() -> None:
    # Tests that a solar model can be solved  
    try_solve('configurations/solar.json', 'configurations/solar_solved.json') 


def test_solve_milk_heater() -> None:
    try_solve('configurations/milk_heater.json', 'configurations/milk_heater_solved.json') 

def test_solve_dynamic_tank() -> None:
    try_solve('configurations/dynamic_tank.json', 'configurations/dynamic_tank_solved.json') 

def test_solve_elimination() -> None:
    # more complex elimination, where there are choices to eliminate and the right pair has to be eliminated
    # ie. because we can't have both enthalpy and temperature fixed at the same time, leaving
    # both flow_mol and flow_mass unfixed (even this is still the correct degrees of freedom)
    # in this specific test case, flow_mass should eliminate flow_mol and constrain at the state block level
    try_solve('configurations/elimination.json', 'configurations/elimination_solved.json')

def test_solve_header() -> None:
    # Tests that a compound separator model can be solved
    try_solve('configurations/header.json', 'configurations/header_solved.json')

def test_initialisation_disabled_branch():
    """
    Test that FlowsheetManager.initialise() skips initialisation and prints the correct message
    when disable_initialization is True.
    """
    schema = FlowsheetSchema(
        id=999,
        dynamic=False,
        time_set=[0],
        property_packages=[],
        unit_models=[],
        arcs=[],
        expressions=[],
        optimizations=[],
        is_rating_mode=False,
        disable_initialization=True,  # Disable initialisation for this test
    )
    manager = FlowsheetManager(schema)
    manager.load()

    f = io.StringIO()
    with redirect_stdout(f):
        manager.initialise()
    output = f.getvalue()

    assert "Initialisation is disabled for this scenario." in output
    assert "initialise_model" not in output


from pyomo.network import Arc
from .flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import PortId


class ArcManager:

    def __init__(self, flowsheet_manager: FlowsheetManager):
        """
        Initializes the arc manager
        """
        self._flowsheet_manager = flowsheet_manager

    def load(self):
        """
        Loads arcs from the schema and adds them to the flowsheet
        """
        schema = self._flowsheet_manager.schema

        for arc_schema in schema.arcs:
            self.add_arc(arc_schema.source, arc_schema.destination)

    def add_arc(self, from_port_id: PortId, to_port_id: PortId):
        """
        Adds an arc between two ports
        """
        port_manager = self._flowsheet_manager.ports
        from_port = port_manager.get_port(from_port_id)
        to_port = port_manager.get_port(to_port_id)
        arc = Arc(source=from_port, destination=to_port)
        self._flowsheet_manager.model.fs.add_component(
            f"arc_{from_port_id}_{to_port_id}", arc
        )



from .methods.load_unit_model import add_unit_model_to_flowsheet
from .methods.adapter_library import UnitModelConstructor, AdapterLibrary
from .flowsheet_manager_type import FlowsheetManager
from ahuora_builder_types import UnitModelId, UnitModelSchema
from idaes.core import UnitModelBlock



class UnitModelManager:
    """
    Manages the unit models in the flowsheet
    """

    def __init__(self, flowsheet_manager: FlowsheetManager) -> None:
        """
        Create a new unit model manager
        """
        self.flowsheet_manager = flowsheet_manager
        self._unit_models: dict[UnitModelId, UnitModelBlock] = {}

    def load(self) -> None:
        """
        load all unit models from the schema into the flowsheet
        """
        schema = self.flowsheet_manager.schema

        for unit_model_def in schema.unit_models:
            unit_model_type = unit_model_def.type
            if unit_model_type not in AdapterLibrary:
                raise Exception(
                    f"Unit model type '{unit_model_type}' is not in the adapter library."
                )
            adapter_constructor: UnitModelConstructor = AdapterLibrary[unit_model_type]
            self.load_from_def(unit_model_def, adapter_constructor)

    def load_from_def(
        self, unit_model_def: UnitModelSchema, adapter_constructor: UnitModelConstructor
    ) -> None:
        """
        Deserialise a unit model from a JSON object into a unit model adapter
        """
        # add the unit model to the flowsheet
        unit_model =  add_unit_model_to_flowsheet(
            unit_model_def, adapter_constructor, self.flowsheet_manager
        )
        
        # store a reference to the adapter
        self._unit_models[unit_model_def.id] = unit_model



# Methods and helper functions to train and use a surrogate valve.
from pyomo.core.base.expression import ScalarExpression
from idaes.core import MaterialBalanceType, ControlVolume0DBlock, declare_process_block_class, EnergyBalanceType, MomentumBalanceType, MaterialBalanceType, useDefault, UnitModelBlockData
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.surrogate.surrogate_block import SurrogateBlock
from idaes.core.util.config import is_physical_parameter_block
from ahuora_builder.methods.expression_parsing import get_property_from_id
from idaes.core.surrogate.pysmo_surrogate import PysmoSurrogate
from io import StringIO
import json
from pyomo.environ import Var, Constraint
from pyomo.environ import (
    Constraint,
    Set,
    Var,
    Suffix,
    units as pyunits,
)

def make_control_volume(unit, name, config):
    if config.dynamic is not False:
        raise ValueError('SurrogateValve does not support dynamics')
    if config.has_holdup is not False:
        raise ValueError('SurrogateValve does not support holdup')

    control_volume = ControlVolume0DBlock(
                                          dynamic=config.dynamic,
                                          has_holdup=config.has_holdup,
                                          property_package=config.property_package,
                                          property_package_args=config.property_package_args)

    # Add the control volume block to the unit
    setattr(unit, name, control_volume)

    control_volume.add_state_blocks(has_phase_equilibrium=config.has_phase_equilibrium)
    # control_volume.add_material_balances(balance_type=config.material_balance_type,
    #                                      has_phase_equilibrium=config.has_phase_equilibrium)
    # control_volume.add_total_enthalpy_balances(has_heat_of_reaction=False, 
    #                                            has_heat_transfer=False, 
    #                                            has_work_transfer=False)
@declare_process_block_class("PySMOModel")
class PySMOModelData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    # Declare all the standard config arguments for the control_volume
    CONFIG.declare("material_balance_type", ConfigValue(default=MaterialBalanceType.componentPhase, domain=In(MaterialBalanceType)))
    CONFIG.declare("energy_balance_type", ConfigValue(default=EnergyBalanceType.enthalpyTotal, domain=In([EnergyBalanceType.enthalpyTotal])))
    CONFIG.declare("momentum_balance_type", ConfigValue(default=MomentumBalanceType.none, domain=In([MomentumBalanceType.none])))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(default=False, domain=In([False])))
    CONFIG.declare("has_pressure_change", ConfigValue(default=False, domain=In([False])))
    CONFIG.declare("property_package", ConfigValue(default=useDefault, domain=is_physical_parameter_block))
    CONFIG.declare("property_package_args", ConfigBlock(implicit=True))
    # no other args need to be declared, we are just hardcoding the valve model.
    CONFIG.declare("model", ConfigValue())
    CONFIG.declare("ids", ConfigValue())
    CONFIG.declare("unitopNames", ConfigValue())
    CONFIG.declare(
            "num_inlets",
            ConfigValue(
                default=False,
                domain=int,
                description="Number of inlets to add",
                doc="Number of inlets to add",
            ),
        )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of outlets to add",
            doc="Number of outlets to add",
        ),
    )
    
    def build(self):
        super(PySMOModelData, self).build()
        self.CONFIG.dynamic = False
        self.CONFIG.has_holdup = False
        # This function handles adding the control volume block to the unit,
        # and addiing the necessary material and energy balances.
        make_control_volume(self, "control_volume", self.CONFIG)

        # self.add_inlet_port()
        # self.add_outlet_port()

        # Defining parameters of state block class
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet

        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        num_inlets = self.config.num_inlets
        self.inlet_list = [ "inlet_" + str(i+1) for i in range(num_inlets) ]    
        self.inlet_set = Set(initialize=self.inlet_list) 
        self.inlet_blocks = []

        for name in self.inlet_list:
            # add properties_inlet_1, properties_inlet2 etc
            state_block = self.config.property_package.state_block_class(
                self.flowsheet().config.time, doc="inlet ml", **tmp_dict
            )
            self.inlet_blocks.append(state_block)
            # Dynamic equivalent to self.properties_inlet_1 = stateblock
            setattr(self,"properties_" + name, state_block)
            # also add the port
            self.add_port(name=name,block=state_block)


        # Add outlet state blocks

        num_outlets = self.config.num_outlets
        self.outlet_list = [ "outlet_" + str(i+1) for i in range(num_outlets) ]
        self.outlet_set = Set(initialize=self.outlet_list)
        self.outlet_blocks = []

        for name in self.outlet_list:
            tmp_dict["defined_state"] = False
            state_block = self.config.property_package.state_block_class(
                self.flowsheet().config.time, doc="outlet ml", **tmp_dict
            )
            self.outlet_blocks.append(state_block)
            setattr(self,"properties_" + name, state_block)
            self.add_port(name=name,block=state_block)
        
        
        
        # Add variables for custom properties        
        names = self.config.unitopNames
        for name in names:
            self.add_component(
                name, Var(self.flowsheet().time, initialize=10)
            )
    
    def initialize(self, *args, **kwargs):
        model_data = self.config.model
        json_str = json.dumps(model_data)
        f = StringIO(json_str)

        model = PysmoSurrogate.load(f)
        ids = self.config.ids
        fs = self.flowsheet()
        # TODO: Make surrogate models work with dynamics. This involves making a surrogate model for each time step.
        # Not sure if idaes has a framework for doing this or not.
        # For now, we are just assuming steady state and time_index=0.
        inputs = [get_property_from_id(fs, i,0) for i in ids["input"]]
        outputs = [get_property_from_id(fs, i,0) for i in ids["output"]]

        self.check_is_expression(inputs)
        self.check_is_expression(outputs)
        
        self.surrogate = SurrogateBlock(concrete=True)
        self.surrogate.build_model(model,input_vars=inputs, output_vars=outputs)
        
    def check_is_expression(self, vars):
        for index, var in enumerate(vars):
            if isinstance(var, ScalarExpression):
                name = f"{var.name}_{index}"
                new_var = Var(self.flowsheet().time, initialize=1)
                self.add_component(name, new_var)
                def constraint_rule(model, t):
                    if var.is_indexed():
                        return new_var[t] == var[t]
                    else:
                        return new_var[t] == var

                self.add_component(f"{name}_constraint", Constraint(self.flowsheet().time, rule=constraint_rule))
                vars[index] = new_var

    def _get_stream_table_contents(self, time_point=0):

        io_dict = {}
        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name) # get a reference to the port
        
        out_dict = {}
        for outlet_name in self.outlet_list:
            out_dict[outlet_name] = getattr(self, outlet_name) # get a reference to the port



from contextvars import ContextVar
import time


class TimingDebugHandler:
    def __init__(self, key: str = "root") -> None:
        self.timing: dict[str, dict] = {}
        self.pointer_stack = [self.timing]
        self.step_into(key)
    
    def current_item(self):
        item = self.current_pointer()
        if len(item) == 0:
            return None
        return item[list(item.keys())[-1]]
    
    def current_pointer(self):
        return self.pointer_stack[-1]
    
    def update_last_timing(self):
        if self.current_item():
            # update the duration of the previous key
            self.current_item()["duration"] = time.time() - self.current_item()["start_time"]

    def add_timing(self, key: str):
        self.update_last_timing()
        self.current_pointer()[key] = {
            "start_time": time.time(),
            "duration": None,  # calculated at the next timing
            "children": {}
        }
    
    def step_into(self, key=None):
        self.add_timing(key)
        self.pointer_stack.append(self.current_item()["children"])
    
    def step_out(self):
        self.update_last_timing()
        self.pointer_stack.pop()

    def close(self):
        # update root duration
        self.step_out()
        self.update_last_timing()
        return self.timing


def start_timing(key: str = "root"):
    """
    start a new timing handler
    """
    handler = TimingDebugHandler(key)
    timing.set(handler)
    return handler


def get_timer():
    """
    get the current timing handler
    """
    return timing.get()


# Create a context variable to store the timing information
timing: ContextVar[TimingDebugHandler | None] = ContextVar("timing", default=None)



### Imports
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination
from pyomo.network import SequentialDecomposition, Port, Arc
from pyomo.core.base.units_container import _PyomoUnit, units 
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
from idaes.core.util.tables import _get_state_from_port
import idaes.logger as idaeslog
from property_packages.build_package import build_package
from idaes.models.unit_models.pressure_changer import Pump
from idaes.models.unit_models.heater import Heater
from ..power_property_package import PowerParameterBlock
from ..bus import Bus
from ..mainDistributionBoard import MDB
from pyomo.environ import value

def test_mdb():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.power_property_package = PowerParameterBlock()
    m.fs.mdb = MDB(
        property_package=m.fs.power_property_package,
        num_inlets=1,
        num_outlets=3,
        )
    m.fs.mdb.inlet_1.power.fix(10 * units.kW)
    m.fs.mdb.priorities[0,'outlet_1'].fix(5* units.kW)
    m.fs.mdb.priorities[0,'outlet_2'].fix(3* units.kW)

    
    # Check that the model is fully specified
    assert degrees_of_freedom(m.fs.mdb) == 0 

    # check that it solves correctly
    solver = SolverFactory("ipopt")
    solver.solve(m)
    assert abs(value(m.fs.mdb.outlet_3.power[0])) <= 2.0005e3 
    
    #Reduce input power and check that it doesn't give power to the last outlet:
    m.fs.mdb.inlet_1.power.fix(8 * units.kW)
    solver = SolverFactory("ipopt")
    solver.solve(m)
    assert abs(value(m.fs.mdb.outlet_3.power[0])) <= 0.0005

    #Increase input power and check that it gives all excess power to the last outlet:
    m.fs.mdb.inlet_1.power.fix(18 * units.kW)
    solver = SolverFactory("ipopt")
    solver.solve(m)
    assert abs(value(m.fs.mdb.outlet_3.power[0])) <= 10.0005e3


from idaes.models.unit_models.heater import HeaterData
from .inverted import add_inverted, initialise_inverted, enable_inverted, disable_inverted
from idaes.core import declare_process_block_class

@declare_process_block_class("CustomCooler")
class CustomCoolerData(HeaterData):
    """
    Custom Cooler model that includes inverted deltaP and Heat Added properties.
    """

    def build(self, *args, **kwargs):
        super().build(*args, **kwargs)

        # add deltaP_inverted as a property
        add_inverted(self, "heat_duty")
        if hasattr(self,"deltaP"):
            add_inverted(self, "deltaP")

    def initialize_build(
        self,*args,**kwargs,
    ):
        initialise_inverted(self, "heat_duty")
        disable_inverted(self, "heat_duty")

        if hasattr(self,"deltaP"):
            initialise_inverted(self, "deltaP")
            disable_inverted(self, "deltaP")

        super().initialize_build(*args, **kwargs)

        enable_inverted(self, "heat_duty")
        if hasattr(self,"deltaP"):
            enable_inverted(self, "deltaP")